########################################################################
# File name: xmltestutils.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest


def element_path(el, upto=None):
    segments = []
    parent = el.getparent()

    while parent != upto:
        similar = list(parent.iterchildren(el.tag))
        index = similar.index(el)
        segments.insert(0, (el.tag, index))
        el = parent
        parent = el.getparent()

    base = "/" + el.tag
    if segments:
        return base + "/" + "/".join(
            "{}[{}]".format(tag, index)
            for tag, index in segments
        )
    return base


class XMLTestCase(unittest.TestCase):
    def assertChildrenEqual(self, tree1, tree2,
                            strict_ordering=False,
                            ignore_surplus_children=False,
                            ignore_surplus_attr=False):
        if not strict_ordering:
            t1_childmap = {}
            for child in tree1:
                t1_childmap.setdefault(child.tag, []).append(child)
            t2_childmap = {}
            for child in tree2:
                t2_childmap.setdefault(child.tag, []).append(child)

            t1_childtags = frozenset(t1_childmap)
            t2_childtags = frozenset(t2_childmap)

            if not ignore_surplus_children or (t1_childtags - t2_childtags):
                self.assertSetEqual(
                    t1_childtags,
                    t2_childtags,
                    "Child tag occurence differences at {}".format(
                        element_path(tree2)))

            for tag, t1_children in t1_childmap.items():
                t2_children = t2_childmap.get(tag, [])
                self.assertLessEqual(
                    len(t1_children),
                    len(t2_children),
                    "Surplus child at {}".format(element_path(tree2))
                )
                if not ignore_surplus_children:
                    self.assertGreaterEqual(
                        len(t1_children),
                        len(t2_children),
                        "Surplus child at {}".format(element_path(tree2))
                    )

                for c1, c2 in zip(t1_children, t2_children):
                    self.assertSubtreeEqual(
                        c1, c2,
                        ignore_surplus_attr=ignore_surplus_attr,
                        ignore_surplus_children=ignore_surplus_children,
                        strict_ordering=strict_ordering)
        else:
            t1_children = list(tree1)
            t2_children = list(tree2)
            self.assertLessEqual(
                len(t1_children),
                len(t2_children),
                "Surplus child at {}".format(element_path(tree2))
            )
            if not ignore_surplus_children:
                self.assertGreaterEqual(
                    len(t1_children),
                    len(t2_children),
                    "Surplus child at {}".format(element_path(tree2))
                )

            for c1, c2 in zip(t1_children, t2_children):
                self.assertSubtreeEqual(
                    c1, c2,
                    ignore_surplus_attr=ignore_surplus_attr,
                    ignore_surplus_children=ignore_surplus_children,
                    strict_ordering=strict_ordering)

    def assertAttributesEqual(self, el1, el2,
                              ignore_surplus_attr=False):
        t1_attrdict = el1.attrib
        t2_attrdict = el2.attrib
        t1_attrs = set(t1_attrdict)
        t2_attrs = set(t2_attrdict)

        if not ignore_surplus_attr or (t1_attrs - t2_attrs):
            self.assertSetEqual(
                t1_attrs,
                t2_attrs,
                "Attribute key differences at {}".format(element_path(el2))
            )

        for attr in t1_attrs:
            self.assertEqual(
                t1_attrdict[attr],
                t2_attrdict[attr],
                "Attribute value difference at {}@{}".format(
                    element_path(el2),
                    attr))

    def _collect_text_parts(self, el):
        parts = [el.text or ""]
        parts.extend(child.tail or "" for child in el)
        return parts

    def assertTextContentEqual(self, el1, el2, join_text_parts=True):
        parts1 = self._collect_text_parts(el1)
        parts2 = self._collect_text_parts(el2)
        if join_text_parts:
            self.assertEqual(
                "".join(parts1),
                "".join(parts2),
                "text mismatch at {}".format(element_path(el2))
            )
        else:
            self.assertSequenceEqual(
                parts1,
                parts2,
                "text mismatch at {}".format(element_path(el2))
            )

    def assertSubtreeEqual(self, tree1, tree2,
                           ignore_surplus_attr=False,
                           ignore_surplus_children=False,
                           strict_ordering=False,
                           join_text_parts=True):
        self.assertEqual(tree1.tag, tree2.tag,
                         "tag mismatch at {}".format(element_path(tree2)))

        self.assertTextContentEqual(tree1, tree2,
                                    join_text_parts=join_text_parts)

        self.assertAttributesEqual(tree1, tree2,
                                   ignore_surplus_attr=ignore_surplus_attr)
        self.assertChildrenEqual(
            tree1, tree2,
            ignore_surplus_children=ignore_surplus_children,
            ignore_surplus_attr=ignore_surplus_attr,
            strict_ordering=strict_ordering)



########################################################################
# File name: __init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.pubsub` --- Publish-Subscribe support (:xep:`0060`)
##################################################################

This subpackage provides client-side support for :xep:`0060` publish-subscribe
services.

.. versionadded:: 0.6

Using Publish-Subscribe
=======================

To start using PubSub services in your application, you have to load the
:class:`.PubSubClient` into the client, using :meth:`~.node.Client.summon`.

.. currentmodule:: aioxmpp

.. autoclass:: PubSubClient

.. currentmodule:: aioxmpp.pubsub

.. class:: Service

   Alias of :class:`.PubSubClient`.

   .. deprecated:: 0.8

      The alias will be removed in 1.0.

.. currentmodule:: aioxmpp.pubsub.xso

XSOs
====

Registering payloads
--------------------

PubSub payloads are must be registered at several places, so there is a
short-hand function to handle that:

.. autofunction:: as_payload_class

Features
--------

.. autoclass:: Feature

Generic namespace
-----------------

The top-level XSO is :class:`Request`. Below that, several different XSOs are
allowed, which are listed below the documentation of :class:`Request` in
alphabetical order.

.. autoclass:: Request

.. autoclass:: Affiliation

.. autoclass:: Affiliations

.. autoclass:: Configure

.. autoclass:: Create

.. autoclass:: Default

.. autoclass:: Item

.. autoclass:: Items

.. autoclass:: Options

.. autoclass:: Publish

.. autoclass:: Retract

.. autoclass:: Subscribe

.. autoclass:: SubscribeOptions

.. autoclass:: Subscription

.. autoclass:: Subscriptions

.. autoclass:: Unsubscribe

Owner namespace
---------------

The top-level XSO is :class:`OwnerRequest`. Below that, several different XSOs
are allowed, which are listed below the documentation of :class:`OwnerRequest`
in alphabetical order.

.. autoclass:: OwnerRequest

.. autoclass:: OwnerAffiliation

.. autoclass:: OwnerAffiliations

.. autoclass:: OwnerConfigure

.. autoclass:: OwnerDefault

.. autoclass:: OwnerDelete

.. autoclass:: OwnerPurge

.. autoclass:: OwnerRedirect

.. autoclass:: OwnerSubscription

.. autoclass:: OwnerSubscriptions

Application-condition error XSOs
--------------------------------

Application-condition XSOs for use in
:attr:`.stanza.Error.application_condition` are also defined for the error
conditions specified by :xep:`0060`. They are listed in alphabetical order
below:

.. autoclass:: ClosedNode()

.. autoclass:: ConfigurationRequired()

.. autoclass:: InvalidJID()

.. autoclass:: InvalidOptions()

.. autoclass:: InvalidPayload()

.. autoclass:: InvalidSubID()

.. autoclass:: ItemForbidden()

.. autoclass:: ItemRequired()

.. autoclass:: JIDRequired()

.. autoclass:: MaxItemsExceeded()

.. autoclass:: MaxNodesExceeded()

.. autoclass:: NodeIDRequired()

.. autoclass:: NotInRosterGroup()

.. autoclass:: NotSubscribed()

.. autoclass:: PayloadTooBig()

.. autoclass:: PayloadRequired()

.. autoclass:: PendingSubscription()

.. autoclass:: PresenceSubscriptionRequired()

.. autoclass:: SubIDRequired()

.. autoclass:: TooManySubscriptions()

.. autoclass:: Unsupported()


"""

from .service import PubSubClient  # NOQA
Service = PubSubClient



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import enum
import unittest

import aioxmpp.xso as xso

from aioxmpp.chatstates import ChatState
from aioxmpp.stanza import Message
from aioxmpp.utils import namespaces


class TestNamespace(unittest.TestCase):

    def test_namespace(self):
        self.assertEqual(namespaces.xep0085,
                         "http://jabber.org/protocol/chatstates")


class TestChatState(unittest.TestCase):

    def test_is_enum(self):
        self.assertTrue(issubclass(ChatState, enum.Enum))

    def test_values(self):
        self.assertEqual(
            ChatState.ACTIVE.value, (namespaces.xep0085, "active"),
        )
        self.assertEqual(
            ChatState.COMPOSING.value, (namespaces.xep0085, "composing"),
        )
        self.assertEqual(
            ChatState.PAUSED.value, (namespaces.xep0085, "paused"),
        )
        self.assertEqual(
            ChatState.INACTIVE.value, (namespaces.xep0085, "inactive"),
        )
        self.assertEqual(
            ChatState.GONE.value, (namespaces.xep0085, "gone"),
        )


class TestMessage(unittest.TestCase):

    def test_xep0085_chatstate(self):
        self.assertIsInstance(
            Message.xep0085_chatstate,
            xso.ChildTag
        )



########################################################################
# File name: test_xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest
import unittest.mock

import multidict

import aioxmpp.shim.xso as shim_xso
import aioxmpp.stanza
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


class TestNamespaces(unittest.TestCase):
    def test_xep0131_shim(self):
        self.assertEqual(
            namespaces.xep0131_shim,
            "http://jabber.org/protocol/shim"
        )


class TestHeader(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            shim_xso.Header,
            xso.XSO,
        ))

    def test_tag(self):
        self.assertEqual(
            shim_xso.Header.TAG,
            (namespaces.xep0131_shim, "header"),
        )

    def test_name(self):
        self.assertIsInstance(
            shim_xso.Header.name,
            xso.Attr
        )
        self.assertEqual(
            shim_xso.Header.name.tag,
            (None, "name")
        )
        self.assertIs(
            shim_xso.Header.name.default,
            xso.NO_DEFAULT
        )

    def test_value(self):
        self.assertIsInstance(
            shim_xso.Header.value,
            xso.Text,
        )

    def test_init(self):
        with self.assertRaises(TypeError):
            shim_xso.Header()

        h = shim_xso.Header("foo", "bar")
        self.assertEqual(h.name, "foo")
        self.assertEqual(h.value, "bar")


class TestHeaderType(unittest.TestCase):
    def test_is_element_type(self):
        self.assertTrue(issubclass(
            shim_xso.HeaderType,
            xso.AbstractElementType
        ))

    def setUp(self):
        self.t = shim_xso.HeaderType()

    def tearDown(self):
        del self.t

    def test_get_xso_types(self):
        self.assertCountEqual(
            self.t.get_xso_types(),
            [shim_xso.Header],
        )

    def test_unpack(self):
        h = unittest.mock.Mock()
        h.name = unittest.mock.sentinel.name
        h.value = unittest.mock.sentinel.value

        self.assertEqual(
            self.t.unpack(h),
            (
                unittest.mock.sentinel.name,
                unittest.mock.sentinel.value,
            )
        )

    def test_pack(self):
        with unittest.mock.patch("aioxmpp.shim.xso.Header") as Header:
            result = self.t.pack(
                (
                    unittest.mock.sentinel.name,
                    unittest.mock.sentinel.value,
                )
            )

        Header.assert_called_with(
            unittest.mock.sentinel.name,
            unittest.mock.sentinel.value,
        )

        self.assertEqual(
            result,
            Header()
        )


class TestHeaders(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            shim_xso.Headers,
            xso.XSO
        ))

    def test_tag(self):
        self.assertEqual(
            shim_xso.Headers.TAG,
            (namespaces.xep0131_shim, "header")
        )

    def test_headers(self):
        self.assertIsInstance(
            shim_xso.Headers.headers,
            xso.ChildValueMultiMap
        )
        self.assertIs(
            shim_xso.Headers.headers.mapping_type,
            multidict.CIMultiDict
        )
        self.assertIsInstance(
            shim_xso.Headers.headers.type_,
            shim_xso.HeaderType,
        )

    def test_headers_attribute_on_Message(self):
        self.assertIsInstance(
            aioxmpp.stanza.Message.xep0131_headers,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.stanza.Message.xep0131_headers._classes,
            {
                shim_xso.Headers,
            }
        )

    def test_headers_attribute_on_Presence(self):
        self.assertIsInstance(
            aioxmpp.stanza.Presence.xep0131_headers,
            xso.Child,
        )
        self.assertSetEqual(
            aioxmpp.stanza.Presence.xep0131_headers._classes,
            {
                shim_xso.Headers,
            }
        )


# foo



########################################################################
# File name: xso.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import enum

import aioxmpp.forms
import aioxmpp.stanza
import aioxmpp.stringprep
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces


namespaces.xep0045_muc = "http://jabber.org/protocol/muc"
namespaces.xep0045_muc_user = "http://jabber.org/protocol/muc#user"
namespaces.xep0045_muc_admin = "http://jabber.org/protocol/muc#admin"
namespaces.xep0045_muc_owner = "http://jabber.org/protocol/muc#owner"
namespaces.xep0249_conference = "jabber:x:conference"


class StatusCode(enum.IntEnum):
    """
    This integer enumeration (see :class:`enum.IntEnum`) is used for the
    status codes defined in :xep:`45`.

    Note that members of this enumeration are equal to their respective integer
    values, making it ideal for backward- and forward-compatible code and a
    replacement for magic numbers.

    .. versionadded:: 0.10

        Before version 0.10, this enum did not exist and the numeric codes
        were used bare. Since this is an :class:`~enum.IntEnum`, it is possible
        to use the named enum members and their numeric codes interchangably.

    .. attribute:: NON_ANONYMOUS
        :annotation: = 100

        Included when entering a room where every user can see every users
        real JID.

    .. attribute:: AFFILIATION_CHANGE
        :annotation: = 101

        Included in out-of-band messages informing about affiliation changes.

    .. attribute:: SHOWING_UNAVAILABLE
        :annotation: = 102

        Inform occupants that room now shows unavailable members.

    .. attribute:: NOT_SHOWING_UNAVAILABLE
        :annotation: = 103

        Inform occupants that room now does not show unavailable members.

    .. attribute:: CONFIG_NON_PRIVACY_RELATED
        :annotation: = 104

        Inform occupants that a non-privacy related configuration change has
        occured.

    .. attribute:: SELF
        :annotation: = 110

        Inform that the stanza refers to the addressee themselves.

    .. attribute:: CONFIG_ROOM_LOGGING
        :annotation: = 170

        Inform that the room is now logged.

    .. attribute:: CONFIG_NO_ROOM_LOGGING
        :annotation: = 171

        Inform that the room is not logged anymore.

    .. attribute:: CONFIG_NON_ANONYMOUS
        :annotation: = 172

        Inform that the room is now not anonymous.

    .. attribute:: CONFIG_SEMI_ANONYMOUS
        :annotation: = 173

        Inform that the room is now semi-anonymous.

    .. attribute:: CREATED
        :annotation: = 201

        Inform that the room was created during the join operation.

    .. attribute:: REMOVED_BANNED
        :annotation: = 301

        Inform that the user was banned from the room.

    .. attribute:: NICKNAME_CHANGE
        :annotation: = 303

        Inform about new nickname.

    .. attribute:: REMOVED_KICKED
        :annotation: = 307

        Inform that the occupant was kicked.

    .. attribute:: REMOVED_AFFILIATION_CHANGE
        :annotation: = 321

        Inform that the occupant was removed from the room due to a change in
        affiliation.

    .. attribute:: REMOVED_NONMEMBER_IN_MEMBERS_ONLY
        :annotation: = 322

        Inform that the occupant was removed from the room because the room was
        changed to members-only and the occupant was not a member.

    .. attribute:: REMOVED_SERVICE_SHUTDOWN
        :annotation: = 332

        Inform that the occupant is being removed because the MUC service is
        being shut down.

    .. attribute:: REMOVED_ERROR
        :annotation: = 333

        Inform that the occupant is being removed because there was an error
        while communicating with them or their server.

    """

    NON_ANONYMOUS = 100
    AFFILIATION_CHANGE = 101
    SHOWING_UNAVAILABLE = 102
    NOT_SHOWING_UNAVAILABLE = 103
    CONFIG_NON_PRIVACY_RELATED = 104
    SELF = 110
    CONFIG_ROOM_LOGGING = 170
    CONFIG_NO_ROOM_LOGGING = 171
    CONFIG_NON_ANONYMOUS = 172
    CONFIG_SEMI_ANONYMOUS = 173
    CREATED = 201
    REMOVED_BANNED = 301
    NICKNAME_CHANGE = 303
    REMOVED_KICKED = 307
    REMOVED_AFFILIATION_CHANGE = 321
    REMOVED_NONMEMBER_IN_MEMBERS_ONLY = 322
    REMOVED_SERVICE_SHUTDOWN = 332
    REMOVED_ERROR = 333


class History(xso.XSO):
    TAG = (namespaces.xep0045_muc, "history")

    maxchars = xso.Attr(
        "maxchars",
        type_=xso.Integer(),
        default=None,
    )

    maxstanzas = xso.Attr(
        "maxstanzas",
        type_=xso.Integer(),
        default=None,
    )

    seconds = xso.Attr(
        "seconds",
        type_=xso.Integer(),
        default=None,
    )

    since = xso.Attr(
        "since",
        type_=xso.DateTime(),
        default=None,
    )

    def __init__(self, *,
                 maxchars=None, maxstanzas=None, seconds=None, since=None):
        super().__init__()
        self.maxchars = maxchars
        self.maxstanzas = maxstanzas
        self.seconds = seconds
        self.since = since


class GenericExt(xso.XSO):
    TAG = (namespaces.xep0045_muc, "x")

    history = xso.Child([History])

    password = xso.ChildText(
        (namespaces.xep0045_muc, "password"),
        default=None
    )


aioxmpp.stanza.Presence.xep0045_muc = xso.Child([
    GenericExt
])

aioxmpp.stanza.Message.xep0045_muc = xso.Child([
    GenericExt
])


class Status(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "status")

    code = xso.Attr(
        "code",
        type_=xso.EnumCDataType(
            StatusCode,
            xso.Integer(),
            allow_coerce=True,
            pass_unknown=True,
        )
    )

    def __init__(self, code):
        super().__init__()
        self.code = code


class StatusCodeList(xso.AbstractElementType):
    def unpack(self, item):
        return item.code

    def pack(self, code):
        item = Status(code)
        return item

    def get_xso_types(self):
        return [Status]


class DestroyNotification(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "destroy")

    reason = xso.ChildText(
        (namespaces.xep0045_muc_user, "reason"),
        default=None
    )

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
        default=None
    )


class Decline(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "decline")

    from_ = xso.Attr(
        "from",
        type_=xso.JID(),
        default=None
    )

    to = xso.Attr(
        "to",
        type_=xso.JID(),
        default=None
    )

    reason = xso.ChildText(
        (namespaces.xep0045_muc_user, "reason"),
        default=None
    )


class Invite(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "invite")

    from_ = xso.Attr(
        "from",
        type_=xso.JID(),
        default=None
    )

    to = xso.Attr(
        "to",
        type_=xso.JID(),
        default=None
    )

    reason = xso.ChildText(
        (namespaces.xep0045_muc_user, "reason"),
        default=None
    )

    password = xso.ChildText(
        (namespaces.xep0045_muc_user, "password"),
        default=None
    )


class ActorBase(xso.XSO):
    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
        default=None,
    )

    nick = xso.Attr(
        "nick",
        type_=xso.String(aioxmpp.stringprep.resourceprep),
        default=None
    )


class ItemBase(xso.XSO):
    affiliation = xso.Attr(
        "affiliation",
        validator=xso.RestrictToSet({
            "admin",
            "member",
            "none",
            "outcast",
            "owner",
            None,
        }),
        validate=xso.ValidateMode.ALWAYS,
        default=None,
    )

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
        default=None,
    )

    nick = xso.Attr(
        "nick",
        type_=xso.String(aioxmpp.stringprep.resourceprep),
        default=None
    )

    role = xso.Attr(
        "role",
        validator=xso.RestrictToSet({
            "moderator",
            "none",
            "participant",
            "visitor",
            None,
        }),
        validate=xso.ValidateMode.ALWAYS,
        default=None,
    )

    def __init__(self,
                 affiliation=None,
                 jid=None,
                 nick=None,
                 role=None,
                 reason=None):
        super().__init__()
        self.affiliation = affiliation
        self.jid = jid
        self.nick = nick
        self.role = role
        self.reason = reason

    @property
    def bare_jid(self):
        """
        Return the bare jid of the item or :data:`None` if no JID is
        given.

        Use this to access the jid unless you really want to know the
        resource. Usually the information given by the resource is
        meaningless (the resource is randomly picked by the server).
        """
        if self.jid:
            return self.jid.bare()
        else:
            return None


class UserActor(ActorBase):
    TAG = (namespaces.xep0045_muc_user, "actor")


class Continue(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "continue")

    thread = xso.Attr(
        "thread",
        type_=aioxmpp.stanza.Thread.identifier.type_,
        default=None,
    )


class UserItem(ItemBase):
    TAG = (namespaces.xep0045_muc_user, "item")

    actor = xso.Child([UserActor])

    continue_ = xso.Child([Continue])

    reason = xso.ChildText(
        (namespaces.xep0045_muc_user, "reason"),
        default=None
    )


class UserExt(xso.XSO):
    TAG = (namespaces.xep0045_muc_user, "x")

    status_codes = xso.ChildValueList(
        StatusCodeList(),
        container_type=set
    )

    destroy = xso.Child([DestroyNotification])

    decline = xso.Child([Decline])

    invites = xso.ChildList([Invite])

    items = xso.ChildList([UserItem])

    password = xso.ChildText(
        (namespaces.xep0045_muc_user, "password"),
        default=None
    )

    def __init__(self,
                 status_codes=[],
                 destroy=None,
                 decline=None,
                 invites=[],
                 items=[],
                 password=None):
        super().__init__()
        self.status_codes.update(status_codes)
        self.destroy = destroy
        self.decline = decline
        self.invites.extend(invites)
        self.items.extend(items)
        self.password = password


aioxmpp.stanza.Presence.xep0045_muc_user = xso.Child([
    UserExt
])

aioxmpp.stanza.Message.xep0045_muc_user = xso.Child([
    UserExt
])


class AdminActor(ActorBase):
    TAG = (namespaces.xep0045_muc_admin, "actor")


class AdminItem(ItemBase):
    TAG = (namespaces.xep0045_muc_admin, "item")

    actor = xso.Child([AdminActor])

    continue_ = xso.Child([Continue])

    reason = xso.ChildText(
        (namespaces.xep0045_muc_admin, "reason"),
        default=None
    )


@aioxmpp.stanza.IQ.as_payload_class
class AdminQuery(xso.XSO):
    TAG = (namespaces.xep0045_muc_admin, "query")

    items = xso.ChildList([AdminItem])

    def __init__(self, *, items=[]):
        super().__init__()
        self.items[:] = items


class DestroyRequest(xso.XSO):
    TAG = (namespaces.xep0045_muc_owner, "destroy")

    reason = xso.ChildText(
        (namespaces.xep0045_muc_owner, "reason"),
        default=None
    )

    password = xso.ChildText(
        (namespaces.xep0045_muc_owner, "password"),
        default=None
    )

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
        default=None
    )


@aioxmpp.stanza.IQ.as_payload_class
class OwnerQuery(xso.XSO):
    TAG = (namespaces.xep0045_muc_owner, "query")

    destroy = xso.Child([DestroyRequest])

    form = xso.Child([aioxmpp.forms.Data])

    def __init__(self, *, form=None, destroy=None):
        super().__init__()
        self.form = form
        self.destroy = destroy


class DirectInvite(xso.XSO):
    TAG = namespaces.xep0249_conference, "x"

    # JEP-0045 v1.19 §6.7 allowed a mediated(!) invitation to contain a
    # (what is now) DirectInvite payload where the reason is included as
    # text (and not as attribute).
    #
    # Some servers still emit this for compatibility. We ignore that.
    _ = xso.Text(default=None)

    jid = xso.Attr(
        "jid",
        type_=xso.JID(),
    )

    reason = xso.Attr(
        "reason",
        default=None,
    )

    password = xso.Attr(
        "password",
        default=None,
    )

    continue_ = xso.Attr(
        "continue",
        type_=xso.Bool(),
        default=False,
    )

    thread = xso.Attr(
        "thread",
        default=None,
    )

    def __init__(self, jid, *,
                 reason=None,
                 password=None,
                 continue_=False,
                 thread=None):
        super().__init__()
        self.jid = jid
        self.reason = reason
        self.password = password
        self.continue_ = continue_
        self.thread = thread


aioxmpp.Message.xep0249_direct_invite = xso.Child([DirectInvite])


class ConfigurationForm(aioxmpp.forms.Form):
    """
    This is a :xep:`4` form template (see :mod:`aioxmpp.forms`) for MUC
    configuration forms.

    The attribute documentation is auto-generated from :xep:`45`; see there for
    details on the semantics of each field.

    .. versionadded:: 0.7
    """

    FORM_TYPE = 'http://jabber.org/protocol/muc#roomconfig'

    maxhistoryfetch = aioxmpp.forms.TextSingle(
        var='muc#maxhistoryfetch',
        label='Maximum Number of History Messages Returned by Room'
    )

    allowpm = aioxmpp.forms.ListSingle(
        var='muc#roomconfig_allowpm',
        label='Roles that May Send Private Messages'
    )

    allowinvites = aioxmpp.forms.Boolean(
        var='muc#roomconfig_allowinvites',
        label='Whether to Allow Occupants to Invite Others'
    )

    changesubject = aioxmpp.forms.Boolean(
        var='muc#roomconfig_changesubject',
        label='Whether to Allow Occupants to Change Subject'
    )

    enablelogging = aioxmpp.forms.Boolean(
        var='muc#roomconfig_enablelogging',
        label='Whether to Enable Public Logging of Room Conversations'
    )

    getmemberlist = aioxmpp.forms.ListMulti(
        var='muc#roomconfig_getmemberlist',
        label='Roles and Affiliations that May Retrieve Member List'
    )

    lang = aioxmpp.forms.TextSingle(
        var='muc#roomconfig_lang',
        label='Natural Language for Room Discussions'
    )

    pubsub = aioxmpp.forms.TextSingle(
        var='muc#roomconfig_pubsub',
        label='XMPP URI of Associated Publish-Subscribe Node'
    )

    maxusers = aioxmpp.forms.ListSingle(
        var='muc#roomconfig_maxusers',
        label='Maximum Number of Room Occupants'
    )

    membersonly = aioxmpp.forms.Boolean(
        var='muc#roomconfig_membersonly',
        label='Whether to Make Room Members-Only'
    )

    moderatedroom = aioxmpp.forms.Boolean(
        var='muc#roomconfig_moderatedroom',
        label='Whether to Make Room Moderated'
    )

    passwordprotectedroom = aioxmpp.forms.Boolean(
        var='muc#roomconfig_passwordprotectedroom',
        label='Whether a Password is Required to Enter'
    )

    persistentroom = aioxmpp.forms.Boolean(
        var='muc#roomconfig_persistentroom',
        label='Whether to Make Room Persistent'
    )

    presencebroadcast = aioxmpp.forms.ListMulti(
        var='muc#roomconfig_presencebroadcast',
        label='Roles for which Presence is Broadcasted'
    )

    publicroom = aioxmpp.forms.Boolean(
        var='muc#roomconfig_publicroom',
        label='Whether to Allow Public Searching for Room'
    )

    roomadmins = aioxmpp.forms.JIDMulti(
        var='muc#roomconfig_roomadmins',
        label='Full List of Room Admins'
    )

    roomdesc = aioxmpp.forms.TextSingle(
        var='muc#roomconfig_roomdesc',
        label='Short Description of Room'
    )

    roomname = aioxmpp.forms.TextSingle(
        var='muc#roomconfig_roomname',
        label='Natural-Language Room Name'
    )

    roomowners = aioxmpp.forms.JIDMulti(
        var='muc#roomconfig_roomowners',
        label='Full List of Room Owners'
    )

    roomsecret = aioxmpp.forms.TextPrivate(
        var='muc#roomconfig_roomsecret',
        label='The Room Password'
    )

    whois = aioxmpp.forms.ListSingle(
        var='muc#roomconfig_whois',
        label='Affiliations that May Discover Real JIDs of Occupants'
    )


class InfoForm(aioxmpp.forms.Form):
    FORM_TYPE = 'http://jabber.org/protocol/muc#roominfo'

    maxhistoryfetch = aioxmpp.forms.TextSingle(
        var='muc#maxhistoryfetch',
        label='Maximum Number of History Messages Returned by Room'
    )

    contactjid = aioxmpp.forms.JIDMulti(
        var='muc#roominfo_contactjid',
        label='Contact Addresses (normally, room owner or owners)'
    )

    description = aioxmpp.forms.TextSingle(
        var='muc#roominfo_description',
        label='Short Description of Room'
    )

    lang = aioxmpp.forms.TextSingle(
        var='muc#roominfo_lang',
        label='Natural Language for Room Discussions'
    )

    ldapgroup = aioxmpp.forms.TextSingle(
        var='muc#roominfo_ldapgroup',
        label='An associated LDAP group that defines room membership; this '
        'should be an LDAP Distinguished Name according to an '
        'implementation-specific or deployment-specific definition of a group.'
    )

    logs = aioxmpp.forms.TextSingle(
        var='muc#roominfo_logs',
        label='URL for Archived Discussion Logs'
    )

    occupants = aioxmpp.forms.TextSingle(
        var='muc#roominfo_occupants',
        label='Current Number of Occupants in Room'
    )

    subject = aioxmpp.forms.TextSingle(
        var='muc#roominfo_subject',
        label='Current Discussion Topic'
    )

    subjectmod = aioxmpp.forms.Boolean(
        var='muc#roominfo_subjectmod',
        label='The room subject can be modified by participants'
    )


class VoiceRequestForm(aioxmpp.forms.Form):
    FORM_TYPE = 'http://jabber.org/protocol/muc#request'

    role = aioxmpp.forms.ListSingle(
        var='muc#role',
        label='Requested role'
    )

    jid = aioxmpp.forms.JIDSingle(
        var='muc#jid',
        label='User ID'
    )

    roomnick = aioxmpp.forms.TextSingle(
        var='muc#roomnick',
        label='Room Nickname'
    )

    request_allow = aioxmpp.forms.Boolean(
        var='muc#request_allow',
        label='Whether to grant voice'
    )



#!/usr/bin/env python3
########################################################################
# File name: buildui.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################

# pyuic is not installed on my system, so here we go with a workaround

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-5", "--qt5",
        dest="version",
        action="store_const",
        const=5,
        default=4)
    parser.add_argument("uifile")
    parser.add_argument("pyfile")

    args = parser.parse_args()

    if args.version == 4:
        import PyQt4.uic as uic
    elif args.version == 5:
        import PyQt5.uic as uic
    else:
        raise ValueError("Invalid version: {}".format(args.version))

    with open(args.pyfile, "w") as f:
        f.write("#pylint: skip-file\n")
        uic.compileUi(args.uifile, f)



########################################################################
# File name: entity_items.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import itertools

import aioxmpp.disco

from framework import Example, exec_example


class EntityItems(Example):
    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        self.argparse.add_argument(
            "target_entity",
            type=jid
        )

        self.argparse.add_argument(
            "target_node",
            default=None,
            nargs="?",
        )

    @asyncio.coroutine
    def run_simple_example(self):
        disco = self.client.summon(aioxmpp.DiscoClient)
        try:
            items = yield from disco.query_items(
                self.args.target_entity,
                node=self.args.target_node,
                timeout=10
            )
        except Exception as exc:
            print("could not get info: ")
            print("{}: {}".format(type(exc).__name__, exc))
            raise

        print("items:")
        for item in items.items:
            print("  jid={}  node={!r}  name={!r}".format(item.jid, item.node, item.name))


if __name__ == "__main__":
    exec_example(EntityItems())



########################################################################
# File name: dispatcher.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import enum

import aioxmpp.callbacks
import aioxmpp.carbons
import aioxmpp.service
import aioxmpp.stream


class MessageSource(enum.Enum):
    STREAM = 0
    CARBONS = 1


class IMDispatcher(aioxmpp.service.Service):
    """
    Dispatches messages, taking into account carbons.

    .. function:: message_filter(message, peer, sent, source)

       A message was received or sent.

       :param message: Message stanza
       :type message: :class:`aioxmpp.Message`
       :param peer: The peer from/to which the stanza was received/sent
       :type peer: :class:`aioxmpp.JID`
       :param sent: Whether the mesasge was sent or received.
       :type sent: :class:`bool`
       :param source: The source of the message.
       :type source: :class:`MessageSource`

       `message` is the message stanza which was sent or received.

       `peer` is the JID of the peer involved in the message. If the message
       was sent, this is the :attr:`~.StanzaBase.to` and otherwise it is the
       :attr:`~.StanzaBase.from_` attribute of the stanza.

       If `sent` is true, the message was sent from this resource *or* another
       resource of the same account, if Message Carbons are enabled.

       `source` indicates how the message was sent or received. It may be one
       of the values of the :class:`MessageSource` enumeration.

    """

    ORDER_AFTER = [
        # we want to be loaded after the SimplePresenceDispatcher to ensure
        # that PresenceClient has updated its data structures before the
        # dispatch_presence handler runs.
        # this helps one-to-one conversations a lot, because they can simply
        # re-use the PresenceClient state
        aioxmpp.dispatcher.SimplePresenceDispatcher,

        aioxmpp.carbons.CarbonsClient,
    ]

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self.message_filter = aioxmpp.callbacks.Filter()
        self.presence_filter = aioxmpp.callbacks.Filter()

    @aioxmpp.service.depsignal(
        aioxmpp.node.Client,
        "before_stream_established")
    @asyncio.coroutine
    def enable_carbons(self, *args):
        carbons = self.dependencies[aioxmpp.carbons.CarbonsClient]
        try:
            yield from carbons.enable()
        except (RuntimeError, aioxmpp.errors.XMPPError):
            self.logger.info(
                "remote server does not support message carbons"
            )
        else:
            self.logger.info(
                "message carbons enabled successfully"
            )
        return True

    @aioxmpp.service.depsignal(
        aioxmpp.stream.StanzaStream,
        "on_message_received")
    def dispatch_message(self, message, *,
                         sent=False,
                         source=MessageSource.STREAM):
        if message.xep0280_received is not None:
            if (message.from_ is not None and
                    message.from_ != self.client.local_jid.bare()):
                return
            message = message.xep0280_received.stanza
            source = MessageSource.CARBONS
        elif message.xep0280_sent is not None:
            if (message.from_ is not None and
                    message.from_ != self.client.local_jid.bare()):
                return
            message = message.xep0280_sent.stanza
            sent = True
            source = MessageSource.CARBONS

        peer = message.to if sent else message.from_

        filtered = self.message_filter.filter(
            message,
            peer,
            sent,
            source,
        )

        if filtered is not None:
            self.logger.debug(
                "message was not processed by any IM handler: %s",
                filtered,
            )

    @aioxmpp.service.depsignal(
        aioxmpp.stream.StanzaStream,
        "on_presence_received")
    def dispatch_presence(self, presence, *, sent=False):
        filtered = self.presence_filter.filter(
            presence,
            presence.from_,
            sent,
        )

        if filtered is not None:
            self.logger.debug(
                "presence was not processed by any IM handler: %s",
                filtered,
            )



########################################################################
# File name: security_layer.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
"""
:mod:`~aioxmpp.security_layer` --- Implementations to negotiate stream security
####################################################################################

This module provides different implementations of the security layer
(TLS+SASL).

These are coupled, as different SASL features might need different TLS features
(such as channel binding or client cert authentication). The preferred method
to construct a :class:`SecurityLayer` is using the :func:`make` function.
:class:`SecurityLayer` objects are needed to establish an XMPP connection,
for example using :class:`aioxmpp.Client`.

.. autofunction:: make

.. autoclass:: PinType

.. autofunction:: tls_with_password_based_authentication(password_provider, [ssl_context_factory], [max_auth_attempts=3])

.. autoclass:: SecurityLayer(ssl_context_factory, certificate_verifier_factory, tls_required, sasl_providers)

.. autofunction:: negotiate_sasl

Certificate verifiers
=====================

To verify the peer certificate provided by the server, different
:class:`CertificateVerifier`s are available:

.. autoclass:: PKIXCertificateVerifier

To implement your own verifiers, see the documentation at the base class for
certificate verifiers:

.. autoclass:: CertificateVerifier

Certificate and key pinning
---------------------------

Often in the XMPP world, we need certificate or public key pinning, as most
XMPP servers do not have certificates trusted by the usual certificate
stores. This module also provide certificate verifiers which can be used for
that purpose, as well as stores for saving the pinned information.

.. autoclass:: PinningPKIXCertificateVerifier

.. autoclass:: CertificatePinStore

.. autoclass:: PublicKeyPinStore

Base classes
^^^^^^^^^^^^

For future expansion or customization, the base classes of the above utilities
can be subclassed and extended:

.. autoclass:: HookablePKIXCertificateVerifier

.. autoclass:: AbstractPinStore

.. _sasl providers:

SASL providers
==============

As elements of the `sasl_providers` argument to :class:`SecurityLayer`,
instances of the following classes can be used:

.. autoclass:: PasswordSASLProvider

.. autoclass:: AnonymousSASLProvider

.. note::

   Patches welcome for additional :class:`SASLProvider` implementations.

Abstract base classes
=====================

For implementation of custom SASL providers, the following base class can be
used:

.. autoclass:: SASLProvider

Deprecated functionality
========================

In pre-0.6 code, you might find use of the following things:

.. autofunction:: security_layer

.. autoclass:: STARTTLSProvider

"""
import abc
import asyncio
import base64
import collections
import enum
import functools
import logging
import ssl

import pyasn1
import pyasn1.codec.der.decoder
import pyasn1.codec.der.encoder
import pyasn1_modules.rfc2459

import OpenSSL.SSL

import aiosasl

from . import errors, sasl, nonza, xso, protocol
from .utils import namespaces


logger = logging.getLogger(__name__)


def extract_python_dict_from_x509(x509):
    """
    Extract a python dictionary similar to the return value of
    :meth:`ssl.SSLSocket.getpeercert` from the given
    :class:`OpenSSL.crypto.X509` `x509` object.

    Note that by far not all attributes are included; only those required to
    use :func:`ssl.match_hostname` are extracted and put in the result.

    In the future, more attributes may be added.
    """
    result = {
        "subject": (
            (("commonName", x509.get_subject().commonName),),
        )
    }

    for ext_idx in range(x509.get_extension_count()):
        ext = x509.get_extension(ext_idx)
        sn = ext.get_short_name()
        if sn != b"subjectAltName":
            continue

        data = pyasn1.codec.der.decoder.decode(
            ext.get_data(),
            asn1Spec=pyasn1_modules.rfc2459.SubjectAltName())[0]
        for name in data:
            dNSName = name.getComponentByPosition(2)
            if dNSName is None:
                continue

            if hasattr(dNSName, "isValue") and not dNSName.isValue:
                continue

            result.setdefault("subjectAltName", []).append(
                ("DNS", str(dNSName))
            )

    return result


def extract_blob(x509):
    """
    Extract an ASN.1 blob from the given :class:`OpenSSL.crypto.X509`
    certificate. Return the resulting :class:`bytes` object.
    """

    return OpenSSL.crypto.dump_certificate(
        OpenSSL.crypto.FILETYPE_ASN1,
        x509)


def blob_to_pyasn1(blob):
    """
    Convert an ASN.1 encoded certificate (such as obtained from
    :func:`extract_blob`) to a :mod:`pyasn1` structure and return the result.
    """

    return pyasn1.codec.der.decoder.decode(
        blob,
        asn1Spec=pyasn1_modules.rfc2459.Certificate()
    )[0]


def extract_pk_blob_from_pyasn1(pyasn1_struct):
    """
    Extract an ASN.1 encoded public key blob from the given :mod:`pyasn1`
    structure (which must represent a certificate).
    """

    pk = pyasn1_struct.getComponentByName(
        "tbsCertificate"
    ).getComponentByName(
        "subjectPublicKeyInfo"
    )

    return pyasn1.codec.der.encoder.encode(pk)


def check_x509_hostname(x509, hostname):
    """
    Check whether the given :class:`OpenSSL.crypto.X509` certificate `x509`
    matches the given `hostname`.

    Return :data:`True` if the name matches and :data:`False` otherwise. This
    uses :func:`ssl.match_hostname` and :func:`extract_python_dict_from_x509`.
    """

    cert_structure = extract_python_dict_from_x509(x509)
    try:
        ssl.match_hostname(cert_structure, hostname)
    except ssl.CertificateError:
        return False
    return True


class CertificateVerifier(metaclass=abc.ABCMeta):
    """
    A certificate verifier hooks into the two mechanisms provided by
    :class:`aioopenssl.STARTTLSTransport` for certificate verification.

    On the one hand, the verify callback provided by
    :class:`OpenSSL.SSL.Context` is used and forwarded to
    :meth:`verify_callback`. On the other hand, the post handshake coroutine is
    set to :meth:`post_handshake`. See the documentation of
    :class:`aioopenssl.STARTTLSTransport` for the semantics of that
    coroutine.

    In addition to these two hooks into the TLS handshake, a third coroutine
    which is called before STARTTLS is intiiated is provided.

    This baseclass provides a bit of boilerplate.
    """

    @asyncio.coroutine
    def pre_handshake(self, metadata, domain, host, port):
        pass

    def setup_context(self, ctx, transport):
        self.transport = transport
        ctx.set_verify(OpenSSL.SSL.VERIFY_PEER, self.verify_callback)

    @abc.abstractmethod
    def verify_callback(self, conn, x509, errno, errdepth, returncode):
        return returncode

    @abc.abstractmethod
    @asyncio.coroutine
    def post_handshake(self, transport):
        pass


class _NullVerifier(CertificateVerifier):
    def setup_context(self, ctx, transport):
        self.transport = transport
        ctx.set_verify(OpenSSL.SSL.VERIFY_NONE, self.verify_callback)

    def verify_callback(self, *args):
        return True

    @asyncio.coroutine
    def post_handshake(self, transport):
        pass


class PKIXCertificateVerifier(CertificateVerifier):
    """
    This verifier enables the default PKIX based verification of certificates
    as implemented by OpenSSL.

    The :meth:`verify_callback` checks that the certificate subject matches the
    domain name of the JID of the connection.
    """

    def verify_callback(self, ctx, x509, errno, errdepth, returncode):
        logger.info("verifying certificate (preverify=%s)", returncode)

        if not returncode:
            logger.warning("certificate verification failed (by OpenSSL)")
            return returncode

        if errdepth == 0:
            hostname = self.transport.get_extra_info("server_hostname")
            if not check_x509_hostname(
                    x509,
                    hostname):
                logger.warning("certificate hostname mismatch "
                               "(doesn’t match for %r)",
                               hostname)
                return False

        return returncode

    def setup_context(self, ctx, transport):
        super().setup_context(ctx, transport)
        ctx.set_default_verify_paths()

    @asyncio.coroutine
    def post_handshake(self, transport):
        pass


class HookablePKIXCertificateVerifier(CertificateVerifier):
    """
    This PKIX-based verifier has several hooks which allow overriding of the
    checking process, for example to implement key or certificate pinning.

    It provides three callbacks:

    * `quick_check` is a synchronous callback (and must be a plain function)
      which is called from :meth:`verify_callback`. It is only called if the
      certificate fails full PKIX verification, and only for certain cases. For
      example, expired certificates do not get a second chance and are rejected
      immediately.

      It is called with the leaf certificate as its only argument. It must
      return :data:`True` if the certificate is known good and should pass the
      verification. If the certificate is known bad and should fail the
      verification immediately, it must return :data:`False`.

      If the certificate is unknown and the check should be deferred to the
      `post_handshake_deferred_failure` callback, :data:`None` must be
      returned.

      Passing :data:`None` to `quick_check` is the same as if a callable passed
      to `quick_check` would return :data:`None` always (i.e. the decision is
      deferred).

    * `post_handshake_deferred_failure` must be a coroutine. It is called after
      the handshake is done but before the STARTTLS negotiation has finished
      and allows the application to take more time to decide on a certificate
      and possibly request user input.

      The coroutine receives the verifier instance as its argument and can make
      use of all the verification attributes to present the user with a
      sensible choice.

      If `post_handshake_deferred_failure` is :data:`None`, the result is
      identical to returning :data:`False` from the callback.

    * `post_handshake_success` is only called if the certificate has passed the
      verification (either because it flawlessly passed by OpenSSL or the
      `quick_check` callback returned :data:`True`).

      You may pass :data:`None` to this argument to disable the callback
      without any further side effects.

    The following attributes are available when the post handshake callbacks
    are called:

    .. attribute:: recorded_errors

       This is a :class:`set` with tuples consisting of a
       :class:`OpenSSL.crypto.X509` instance, an OpenSSL error number and the
       depth of the certificate in the verification chain (0 is the leaf
       certificate).

       It is a collection of all errors which were passed into
       :meth:`verify_callback` by OpenSSL.

    .. attribute:: hostname_matches

       This is :data:`True` if the host name in the leaf certificate matches
       the domain part of the JID for which we are connecting (i.e. the usual
       server name check).

    .. attribute:: leaf_x509

       The :class:`OpenSSL.crypto.X509` object which represents the leaf
       certificate.

    """

    # these are the errors for which we allow pinning the certificate
    _DEFERRABLE_ERRORS = {
        (20, None),  # issuer certificate not available locally
        (19, None),  # self-signed cert in chain
        (18, 0),     # depth-zero self-signed cert
        (27, 0),     # cert untrusted
        (21, 0),     # leaf certificate not signed ...
    }

    def __init__(self,
                 quick_check,
                 post_handshake_deferred_failure,
                 post_handshake_success):
        self._quick_check = quick_check
        self._post_handshake_success = post_handshake_success
        self._post_handshake_deferred_failure = post_handshake_deferred_failure

        self.recorded_errors = set()
        self.deferred = True
        self.hostname_matches = False
        self.leaf_x509 = None

    def verify_callback(self, ctx, x509, errno, depth, preverify):
        if errno != 0 and errno != 21:
            self.recorded_errors.add((x509, errno, depth))
            return True

        if depth == 0:
            if errno != 0:
                logger.debug(
                    "unsigned certificate; this is odd to say the least"
                )
                self.recorded_errors.add((x509, errno, depth))
            hostname = self.transport.get_extra_info("server_hostname")
            self.hostname_matches = check_x509_hostname(x509, hostname)
            self.leaf_x509 = x509
            return self.verify_recorded(x509, self.recorded_errors)

        return True

    def verify_recorded(self, leaf_x509, records):
        self.deferred = False

        if not records:
            return True

        hostname = self.transport.get_extra_info("server_hostname")
        self.hostname_matches = check_x509_hostname(leaf_x509, hostname)

        for x509, errno, depth in records:
            if     ((errno, depth) not in self._DEFERRABLE_ERRORS and
                    (errno, None) not in self._DEFERRABLE_ERRORS):
                logger.warning("non-deferrable certificate error: "
                               "depth=%d, errno=%d",
                               depth, errno)
                return False

        if self._quick_check is not None:
            result = self._quick_check(leaf_x509)
            logger.debug("certificate quick-check returned %r", result)
        else:
            result = None
            logger.debug("no certificate quick-check")

        if result is None:
            self.deferred = True

        return result is not False

    @asyncio.coroutine
    def post_handshake(self, transport):
        if self.deferred:
            if self._post_handshake_deferred_failure is not None:
                result = yield from self._post_handshake_deferred_failure(self)
            else:
                result = False

            if not result:
                raise errors.TLSFailure("certificate verification failed")
        else:
            if self._post_handshake_success is not None:
                yield from self._post_handshake_success()


class AbstractPinStore(metaclass=abc.ABCMeta):
    """
    This is the abstract base class for both :class:`PublicKeyPinStore` and
    :class:`CerificatePinStore`. The interface for both types of pinning is
    identical; the only difference is in which information is stored.

    .. automethod:: pin

    .. automethod:: query

    .. automethod:: get_pinned_for_host

    .. automethod:: export_to_json

    .. automethod:: import_from_json

    For subclasses:

    .. automethod:: _encode_key

    .. automethod:: _decode_key

    .. automethod:: _x509_key


    """

    def __init__(self):
        self._storage = {}

    @abc.abstractmethod
    def _x509_key(self, key):
        """
        Return a hashable value which identifies the given `x509` certificate
        for the purposes of the key store. See the implementations
        :meth:`PublicKeyPinStore._x509_key` and
        :meth:`CertificatePinStore._x509_key` for details on what is stored for
        the respective subclasses.

        This method is abstract and must be implemented in subclasses.
        """

    def _encode_key(self, key):
        """
        Encode the `key` (which has previously been obtained from
        :meth:`_x509_key`) into a string which is both JSON compatible and can
        be used as XML text (which means that it must not contain control
        characters, for example).

        The method is called by :meth:`export_to_json`. The default
        implementation returns `key`.
        """
        return key

    def _decode_key(self, obj):
        """
        Decode the `obj` into a key which is compatible to the values returned
        by :meth:`_x509_key`.

        The method is called by :meth:`import_from_json`. The default
        implementation returns `obj`.
        """
        return obj

    def pin(self, hostname, x509):
        """
        Pin an :class:`OpenSSL.crypto.X509` object `x509` for use with the
        given `hostname`. Which information exactly is used to identify the
        certificate depends :meth:`_x509_key`.
        """

        key = self._x509_key(x509)
        self._storage.setdefault(hostname, set()).add(key)

    def query(self, hostname, x509):
        """
        Return true if the given :class:`OpenSSL.crypto.X509` object `x509` has
        previously been pinned for use with the given `hostname` and
        :data:`None` otherwise.

        Returning :data:`None` allows this method to be used with
        :class:`PinningPKIXCertificateVerifier`.
        """

        key = self._x509_key(x509)
        try:
            pins = self._storage[hostname]
        except KeyError:
            return None

        if key in pins:
            return True

        return None

    def get_pinned_for_host(self, hostname):
        """
        Return the set of hashable values which are used to identify the X.509
        certificates which are accepted for the given `hostname`.

        If no values have previously been pinned, this returns the empty set.
        """
        try:
            return frozenset(self._storage[hostname])
        except KeyError:
            return frozenset()

    def export_to_json(self):
        """
        Return a JSON dictionary which contains all the pins stored in this
        store.
        """

        return {
            hostname: sorted(self._encode_key(key) for key in pins)
            for hostname, pins in self._storage.items()
        }

    def import_from_json(self, data, *, override=False):
        """
        Import a JSON dictionary which must have the same format as exported by
        :meth:`export`.

        If *override* is true, the existing data in the pin store will be
        overriden with the data from `data`. Otherwise, the `data` will be
        merged into the store.
        """

        if override:
            self._storage = {
                hostname: set(self._decode_key(key) for key in pins)
                for hostname, pins in data.items()
            }
            return

        for hostname, pins in data.items():
            existing_pins = self._storage.setdefault(hostname, set())
            existing_pins.update(self._decode_key(key) for key in pins)


class PublicKeyPinStore(AbstractPinStore):
    """
    This pin store stores the public keys of the X.509 objects which are passed
    to its :meth:`pin` method.
    """

    def _x509_key(self, x509):
        blob = extract_blob(x509)
        pyasn1_struct = blob_to_pyasn1(blob)
        return extract_pk_blob_from_pyasn1(pyasn1_struct)

    def _encode_key(self, key):
        return base64.b64encode(key).decode("ascii")

    def _decode_key(self, obj):
        return base64.b64decode(obj.encode("ascii"))


class CertificatePinStore(AbstractPinStore):
    """
    This pin store stores the whole certificates which are passed to its
    :meth:`pin` method.
    """

    def _x509_key(self, x509):
        return extract_blob(x509)

    def _encode_key(self, key):
        return base64.b64encode(key).decode("ascii")

    def _decode_key(self, obj):
        return base64.b64decode(obj.encode("ascii"))


class PinningPKIXCertificateVerifier(HookablePKIXCertificateVerifier):
    """
    The :class:`PinningPKIXCertificateVerifier` is a subclass of the
    :class:`HookablePKIXCertificateVerifier` which uses the hooks to implement
    certificate or public key pinning.

    It does not store the pins itself. Instead, the user must pass a callable
    to the `query_pin` argument. That callable will be called with two
    arguments: the `servername` and the `x509`. The `x509` is a
    :class:`OpenSSL.crypto.X509` instance, which is the leaf certificate which
    attempts to identify the host. The `servername` is the name of the server
    we try to connect to (the identifying name, like the domain part of the
    JID). The callable must return :data:`True` (to accept the certificate),
    :data:`False` (to reject the certificate) or :data:`None` (to defer the
    decision to the `post_handshake_deferred_failure` callback). `query_pin`
    must not block; if it needs to do blocking operations, it should defer.

    The other two arguments are coroutines with semantics identical to those of
    the same-named arguments in :class:`HookablePKIXCertificateVerifier`.

    .. seealso::

       :meth:`AbstractPinStore.query` is a method which can be passed as
       `query_pin` callback.

    """

    def __init__(self,
                 query_pin,
                 post_handshake_deferred_failure,
                 post_handshake_success=None):
        super().__init__(
            self._quick_check_query_pin,
            post_handshake_deferred_failure,
            post_handshake_success
        )

        self._query_pin = query_pin

    def _quick_check_query_pin(self, leaf_x509):
        hostname = self.transport.get_extra_info("server_hostname")
        is_pinned = self._query_pin(hostname, leaf_x509)
        if not is_pinned:
            logger.debug(
                "certificate for %r does not appear in pin store",
                hostname,
            )
        return is_pinned


class ErrorRecordingVerifier(CertificateVerifier):
    def __init__(self):
        super().__init__()
        self._errors = []

    def _record_verify_info(self, x509, errno, depth):
        self._errors.append((x509, errno, depth))

    def verify_callback(self, x509, errno, depth, returncode):
        self._record_verify_info(x509, errno, depth)
        return True

    @asyncio.coroutine
    def post_handshake(self, transport):
        if self._errors:
            raise errors.TLSFailure(
                "Peer certificate verification failure: {}".format(
                    ", ".join(map(str, self._errors))))


class SASLMechanism(xso.XSO):
    TAG = (namespaces.sasl, "mechanism")

    name = xso.Text()

    def __init__(self, name=None):
        super().__init__()
        self.name = name


@nonza.StreamFeatures.as_feature_class
class SASLMechanisms(xso.XSO):
    TAG = (namespaces.sasl, "mechanisms")

    mechanisms = xso.ChildList([SASLMechanism])

    def get_mechanism_list(self):
        return [
            mechanism.name
            for mechanism in self.mechanisms
        ]


class SASLProvider:
    """
    Base class to implement a SASL provider.

    SASL providers are used in :class:`SecurityLayer` to authenticate the local
    user with a service. The credentials required depend on the specific SASL
    provider, and it is recommended to acquire means to get these credentials
    via constructor parameters (see for example :class:`PasswordSASLProvider`).

    The following methods must be implemented by subclasses:

    .. automethod:: execute

    The following methods are intended to be re-used by subclasses:

    .. automethod:: _execute

    .. automethod:: _find_supported
    """

    def _find_supported(self, features, mechanism_classes):
        """
        Find the first mechansim class which supports a mechanism announced in
        the given stream features.

        :param features: Current XMPP stream features
        :type features: :class:`~.nonza.StreamFeatures`
        :param mechanism_classes: SASL mechanism classes to use
        :type mechanism_classes: iterable of :class:`SASLMechanism`
                                 sub\ *classes*
        :raises aioxmpp.errors.SASLUnavailable: if the peer does not announce
                                                SASL support
        :return: the :class:`SASLMechanism` subclass to use and a token
        :rtype: pair

        Return a supported SASL mechanism class, by looking the given
        stream features `features`.

        If no matching mechanism is found, ``(None, None)`` is
        returned. Otherwise, a pair consisting of the mechanism class and the
        value returned by the respective
        :meth:`~.sasl.SASLMechanism.any_supported` method is returned. The
        latter is an opaque token which must be passed to the `token` argument
        of :meth:`_execute` or :meth:`aiosasl.SASLMechanism.authenticate`.
        """

        try:
            mechanisms = features[SASLMechanisms]
        except KeyError:
            logger.error("No sasl mechanisms: %r", list(features))
            raise errors.SASLUnavailable(
                "Remote side does not support SASL") from None

        remote_mechanism_list = mechanisms.get_mechanism_list()

        for our_mechanism in mechanism_classes:
            token = our_mechanism.any_supported(remote_mechanism_list)
            if token is not None:
                return our_mechanism, token

        return None, None

    AUTHENTICATION_FAILURES = {
        "credentials-expired",
        "account-disabled",
        "invalid-authzid",
        "not-authorized",
        "temporary-auth-failure",
    }

    MECHANISM_REJECTED_FAILURES = {
        "invalid-mechanism",
        "mechanism-too-weak",
        "encryption-required",
    }

    @asyncio.coroutine
    def _execute(self, intf, mechanism, token):
        """
        Execute a SASL authentication process.

        :param intf: SASL interface to use
        :type intf: :class:`~.sasl.SASLXMPPInterface`
        :param mechanism: SASL mechanism to use
        :type mechanism: :class:`aiosasl.SASLMechanism`
        :param token: The opaque token argument for the mechanism
        :type token: not :data:`None`
        :raises aiosasl.AuthenticationFailure: if authentication failed due to
                                               bad credentials
        :raises aiosasl.SASLFailure: on other SASL error conditions (such as
                                     protocol violations)
        :return: true if authentication succeeded, false if the mechanism has
                 to be disabled
        :rtype: :class:`bool`

        This executes the SASL authentication process. The more specific
        exceptions are generated by inspecting the
        :attr:`aiosasl.SASLFailure.opaque_error` on exceptinos raised from the
        :class:`~.sasl.SASLXMPPInterface`. Other :class:`aiosasl.SASLFailure`
        exceptions are re-raised without modification.
        """
        sm = aiosasl.SASLStateMachine(intf)
        try:
            yield from mechanism.authenticate(sm, token)
            return True
        except aiosasl.SASLFailure as err:
            if err.opaque_error in self.AUTHENTICATION_FAILURES:
                raise aiosasl.AuthenticationFailure(
                    opaque_error=err.opaque_error,
                    text=err.text)
            elif err.opaque_error in self.MECHANISM_REJECTED_FAILURES:
                return False
            raise

    @abc.abstractmethod
    @asyncio.coroutine
    def execute(self,
                client_jid,
                features,
                xmlstream,
                tls_transport):
        """
        Perform SASL negotiation.

        :param client_jid: The JID the client attempts to authenticate for
        :type client_jid: :class:`aioxmpp.JID`
        :param features: Current stream features nonza
        :type features: :class:`~.nonza.StreamFeatures`
        :param xmlstream: The XML stream to authenticate over
        :type xmlstream: :class:`~.protocol.XMLStream`
        :param tls_transport: The TLS transport or :data:`None` if no TLS has
                              been negotiated
        :type tls_transport: :class:`asyncio.Transport` or :data:`None`
        :raise aiosasl.AuthenticationFailure: if authentication failed due to
                                              bad credentials
        :raise aiosasl.SASLFailure: on other SASL-related errors
        :return: true if the negotiation was successful, false if no common
                 mechanisms could be found or all mechanisms failed for reasons
                 unrelated to the credentials themselves.
        :rtype: :class:`bool`

        The implementation depends on the specific :class:`SASLProvider`
        subclass in use.

        This coroutine returns :data:`True` if the negotiation was
        successful. If no common mechanisms could be found, :data:`False` is
        returned. This is useful to chain several SASL providers (e.g. a
        provider supporting ``EXTERNAL`` in front of password-based providers).

        Any other error case, such as no SASL support on the remote side or
        authentication failure results in an :class:`aiosasl.SASLFailure`
        exception to be raised.
        """


class PasswordSASLProvider(SASLProvider):
    """
    Perform password-based SASL authentication.

    :param password_provider: A coroutine function returning the password to
                              authenticate with.
    :type password_provider: coroutine function
    :param max_auth_attempts: Maximum number of authentication attempts with a
                              single mechansim.
    :type max_auth_attempts: positive :class:`int`

    `password_provider` must be a coroutine taking two arguments, a JID and an
    integer number. The first argument is the JID which is trying to
    authenticate and the second argument is the number of the authentication
    attempt, starting at 0. On each attempt, the number is increased, up to
    `max_auth_attempts`\ -1. If the coroutine returns :data:`None`, the
    authentication process is aborted. If the number of attempts are exceeded,
    the authentication process is also aborted. In both cases, an
    :class:`aiosasl.AuthenticationFailure` error will be raised.

    The SASL mechanisms used depend on whether TLS has been negotiated
    successfully before. In any case, :class:`aiosasl.SCRAM` is used. If TLS has
    been negotiated, :class:`aiosasl.PLAIN` is also supported.

    .. seealso::

       :class:`SASLProvider`
          for the public interface of this class.
    """

    def __init__(self, password_provider, *,
                 max_auth_attempts=3, **kwargs):
        super().__init__(**kwargs)
        self._password_provider = password_provider
        self._max_auth_attempts = max_auth_attempts

    @asyncio.coroutine
    def execute(self,
                client_jid,
                features,
                xmlstream,
                tls_transport):
        client_jid = client_jid.bare()

        password_signalled_abort = False
        nattempt = 0
        cached_credentials = None

        @asyncio.coroutine
        def credential_provider():
            nonlocal password_signalled_abort, nattempt, cached_credentials
            if cached_credentials is not None:
                return client_jid.localpart, cached_credentials

            password = yield from self._password_provider(
                client_jid, nattempt)
            if password is None:
                password_signalled_abort = True
                raise aiosasl.AuthenticationFailure(
                    "user intervention",
                    text="authentication aborted by user")
            cached_credentials = password
            return client_jid.localpart, password

        classes = [
            aiosasl.SCRAM
        ]
        if tls_transport is not None:
            classes.append(aiosasl.PLAIN)

        intf = sasl.SASLXMPPInterface(xmlstream)
        while classes:
            # go over all mechanisms available. some errors disable a mechanism
            # (like encryption-required or mechansim-too-weak)
            mechanism_class, token = self._find_supported(features, classes)
            if mechanism_class is None:
                return False

            mechanism = mechanism_class(credential_provider)
            last_auth_error = None
            for nattempt in range(self._max_auth_attempts):
                try:
                    mechanism_worked = yield from self._execute(
                        intf, mechanism, token)
                except (ValueError, aiosasl.AuthenticationFailure) as err:
                    if password_signalled_abort:
                        # immediately re-raise
                        raise
                    last_auth_error = err
                    # allow the user to re-try
                    cached_credentials = None
                    continue
                else:
                    break
            else:
                raise last_auth_error

            if mechanism_worked:
                return True
            classes.remove(mechanism_class)

        return False


class AnonymousSASLProvider(SASLProvider):
    """
    Perform the ``ANONYMOUS`` SASL mechanism (:rfc:`4505`).

    :param token: The trace token for the ``ANONYMOUS`` mechanism
    :type token: :class:`str`

    `token` SHOULD be the empty string in the XMPP context (see :xep:`175`).

    .. seealso::

       :class:`SASLProvider`
          for the public interface of this class.

    .. warning::

       Take the security and privacy considerations from :rfc:`4505` (which
       specifies the ANONYMOUS SASL mechanism) and :xep:`175` (which discusses
       the ANONYMOUS SASL mechanism in the XMPP context) into account before
       using this provider.

    .. note::

       This class requires :class:`aiosasl.ANONYMOUS`, which is available with
       :mod:`aiosasl` 0.3 or newer. If :class:`aiosasl.ANONYMOUS` is not
       provided, this class is replaced with :data:`None`.

    .. versionadded:: 0.8
    """

    def __init__(self, token):
        super().__init__()
        self._token = token

    @asyncio.coroutine
    def execute(self, client_jid, features, xmlstream, tls_transport):
        mechanism_class, token = self._find_supported(
            features,
            [aiosasl.ANONYMOUS]
        )

        if mechanism_class is None:
            return False

        intf = sasl.SASLXMPPInterface(xmlstream)
        mechanism = aiosasl.ANONYMOUS(self._token)

        return (yield from self._execute(
            intf,
            mechanism,
            token,
        ))


if not hasattr(aiosasl, "ANONYMOUS"):
    AnonymousSASLProvider = None  # NOQA


class SecurityLayer(collections.namedtuple(
        "SecurityLayer",
        [
            "ssl_context_factory",
            "certificate_verifier_factory",
            "tls_required",
            "sasl_providers",
        ])):
    """
    A security layer defines the security properties used for an XML stream.
    This includes TLS settings and SASL providers. The arguments are used to
    initialise the attributes of the same name.

    :class:`SecurityLayer` instances are required to construct a
    :class:`aioxmpp.Client`.

    .. versionadded:: 0.6

    .. seealso::

       :func:`make`
          A powerful function which can be used to create a configured
          :class:`SecurityLayer` instance.

    .. attribute:: ssl_context_factory

       This is a callable returning a :class:`OpenSSL.SSL.Context` instance
       which is to be used for any SSL operations for the connection.

       The :class:`OpenSSL.SSL.Context` instances should not be resued between
       connection attempts, as the certificate verifiers may set options which
       cannot be disabled anymore.

    .. attribute:: certificate_verifier_factory

       This is a callable which returns a fresh
       :class:`CertificateVerifier` on each call (it must be a fresh instance
       since :class:`CertificateVerifier` objects are allowed to keep state and
       :class:`SecurityLayer` objects are reusable between connection
       attempts).

    .. attribute:: tls_required

       A boolean which indicates whether TLS is required. If it is set to true,
       connectors (see :mod:`aioxmpp.connector`) will abort the connection if
       TLS (or something equivalent) is not available on the transport.

       .. note::

           Disabling this makes your application vulnerable to STARTTLS
           stripping attacks.

    .. attribute:: sasl_providers

       A sequence of :class:`SASLProvider` instances. As SASL providers are
       stateless, it is not necessary to create new providers for each
       connection.
    """


def default_verify_callback(conn, x509, errno, errdepth, returncode):
    return errno == 0


def default_ssl_context():
    """
    Return a sensibly configured :class:`OpenSSL.SSL.Context` context.

    The context has SSLv2 and SSLv3 disabled, and supports TLS 1.0+ (depending
    on the version of the SSL library).

    Tries to negotiate an XMPP c2s connection via ALPN (:rfc:`7301`).
    """

    ctx = OpenSSL.SSL.Context(OpenSSL.SSL.SSLv23_METHOD)
    ctx.set_options(OpenSSL.SSL.OP_NO_SSLv2 | OpenSSL.SSL.OP_NO_SSLv3)
    ctx.set_verify(OpenSSL.SSL.VERIFY_PEER, default_verify_callback)
    return ctx


@asyncio.coroutine
def negotiate_sasl(transport, xmlstream,
                   sasl_providers,
                   negotiation_timeout,
                   jid, features):
    """
    Perform SASL authentication on the given :class:`.protocol.XMLStream`
    `stream`. `transport` must be the :class:`asyncio.Transport` over which the
    `stream` runs. It is used to detect whether TLS is used and may be required
    by some SASL mechanisms.

    `sasl_providers` must be an iterable of :class:`SASLProvider` objects. They
    will be tried in iteration order to authenticate against the server. If one
    of the `sasl_providers` fails with a :class:`aiosasl.AuthenticationFailure`
    exception, the other providers are still tried; only if all providers fail,
    the last :class:`aiosasl.AuthenticationFailure` exception is re-raised.

    If no mechanism was able to authenticate but not due to authentication
    failures (other failures include no matching mechanism on the server side),
    :class:`aiosasl.SASLUnavailable` is raised.

    Return the :class:`.nonza.StreamFeatures` obtained after resetting the
    stream after successful SASL authentication.

    .. versionadded:: 0.6

    .. deprecated:: 0.10

        The `negotiation_timeout` argument is ignored. The timeout is
        controlled using the :attr:`~.XMLStream.deadtime_hard_limit` timeout
        of the stream.

        The argument will be removed in version 1.0. To prepare for this, please
        pass `jid` and `features` as keyword arguments.
    """

    if not transport.get_extra_info("sslcontext"):
        transport = None

    last_auth_error = None
    for sasl_provider in sasl_providers:
        try:
            result = yield from sasl_provider.execute(
                jid, features, xmlstream, transport)
        except ValueError as err:
            raise errors.StreamNegotiationFailure(
                "invalid credentials: {}".format(err)
            ) from err
        except aiosasl.AuthenticationFailure as err:
            last_auth_error = err
            continue

        if result:
            features = yield from protocol.reset_stream_and_get_features(
                xmlstream
            )
            break
    else:
        if last_auth_error:
            raise last_auth_error
        else:
            raise errors.SASLUnavailable("No common mechanisms")

    return features


class STARTTLSProvider:
    """
    .. deprecated:: 0.6

       Do **not** use this. This is a shim class which provides
       backward-compatibility for versions older than 0.6.
    """

    def __init__(self, ssl_context_factory,
                 certificate_verifier_factory=PKIXCertificateVerifier,
                 *,
                 require_starttls=True):
        self.ssl_context_factory = ssl_context_factory
        self.certificate_verifier_factory = certificate_verifier_factory
        self.tls_required = require_starttls


def security_layer(tls_provider, sasl_providers):
    """
    .. deprecated:: 0.6

       Replaced by :class:`SecurityLayer`.

    Return a configured :class:`SecurityLayer`. `tls_provider` must be a
    :class:`STARTTLSProvider`.

    The return value can be passed to the constructor of
    :class:`~.node.Client`.

    Some very basic checking on the input is also performed.
    """

    sasl_providers = tuple(sasl_providers)

    if not sasl_providers:
        raise ValueError("At least one SASL provider must be given.")
    for sasl_provider in sasl_providers:
        sasl_provider.execute  # check that sasl_provider has execute method

    result = SecurityLayer(
        tls_provider.ssl_context_factory,
        tls_provider.certificate_verifier_factory,
        tls_provider.tls_required,
        sasl_providers
    )

    return result


def tls_with_password_based_authentication(
        password_provider,
        ssl_context_factory=default_ssl_context,
        max_auth_attempts=3,
        certificate_verifier_factory=PKIXCertificateVerifier):
    """
    Produce a commonly used :class:`SecurityLayer`, which uses TLS and
    password-based SASL authentication. If `ssl_context_factory` is not
    provided, an SSL context with TLSv1+ is used.

    `password_provider` must be a coroutine which is called with the jid
    as first and the number of attempt as second argument. It must return the
    password to us, or :data:`None` to abort.

    Return a :class:`SecurityLayer` instance.

    .. deprecated:: 0.7

       Use :func:`make` instead.
    """

    tls_kwargs = {}
    if certificate_verifier_factory is not None:
        tls_kwargs["certificate_verifier_factory"] = \
            certificate_verifier_factory

    return SecurityLayer(
        ssl_context_factory,
        certificate_verifier_factory,
        True,
        (
            PasswordSASLProvider(
                password_provider,
                max_auth_attempts=max_auth_attempts),
        )
    )


class PinType(enum.Enum):
    """
    Enumeration to control which pinning is used by :meth:`make`.

    .. attribute:: PUBLIC_KEY

       Public keys are stored in the pin store. :class:`PublicKeyPinStore` is
       used.

    .. attribute:: CERTIFICATE

       Whole certificates are stored in the pin store.
       :class:`CertificatePinStore` is used.
    """

    PUBLIC_KEY = 0
    CERTIFICATE = 1


def make(
        password_provider,
        *,
        pin_store=None,
        pin_type=PinType.PUBLIC_KEY,
        post_handshake_deferred_failure=None,
        anonymous=False,
        no_verify=False):
    """
    Construct a :class:`SecurityLayer`. Depending on the arguments passed,
    different features are enabled or disabled.

    :param password_provider: Source for the password to authenticate with.
    :type password_provider: :class:`str` or coroutine
    :param pin_store: Enable use of certificate pin store: if it is a
        :class:`dict`, a new pin store is created (see `pin_type` argument) and
        filled with the data from the dict. Otherwise, the given pin store is
        used.
    :type pin_store: :class:`dict` (compatible to
        :meth:`~AbstractPinStore.import_from_json`) or
        :class:`AbstractPinStore`
    :param pin_type: Type of pin store to use with a dict passed to `pin_store`
        (ignored if no dict is passed to `pin_store`).
    :type pin_type: :class:`PinType`
    :param post_handshake_deferred_failure: Coroutine to call when using pin
        store and the certificate is not in the pin store and fails PKI
        verification.
    :type post_handshake_deferred_failure: coroutine
    :param anonymous: trace token for SASL ANONYMOUS (:rfc:`4505`), enables
        ANONYMOUS authentication
    :type anonymous: :class:`str` or :data:`False`
    :param no_verify: *Disable* all certificate verification. Usage is
        **strongly discouraged** outside controlled test environments. See
        below for alternatives.
    :type no_verify: :class:`bool`
    :raise RuntimeError: if `anonymous` is a :class:`str` and the version of
        :mod:`aiosasl` in use does not provide :class:`aiosasl.ANONYMOUS`
    :return: A new :class:`SecurityLayer` instance configured as per the
        arguments.

    `password_provider` must either be a coroutine or a :class:`str`. As a
    coroutine, it is called during authentication with the JID we are trying to
    authenticate against as the first, and the sequence number of the
    authentication attempt as second argument. The sequence number starts at 0.
    The coroutine is expected to return :data:`None` or a password. See
    :class:`PasswordSASLProvider` for details. If it is a :class:`str`, a
    coroutine which returns the string on the first and :data:`None` on
    subsequent attempts is created and used.

    If `pin_store` is not :data:`None`, :class:`PinningPKIXCertificateVerifier`
    is used instead of the default :class:`PKIXCertificateVerifier`. The
    `pin_store` argument determines the pinned certificates: if it is a
    dictionary, a :class:`AbstractPinStore` according to the :class:`PinType`
    passed as `pin_type` argument is created and initialised with the data from
    the dictionary using its :meth:`~AbstractPinStore.import_from_json` method.
    Otherwise, `pin_store` must be a :class:`AbstractPinStore` instance which
    is passed to the verifier.

    `post_handshake_deferred_callback` is used only if `pin_store` is not
    :data:`None`. It is passed to the equally-named argument of
    :class:`PinningPKIXCertificateVerifier`, see the documentation there for
    details on the semantics. If `post_handshake_deferred_callback` is
    :data:`None` while `pin_store` is not, a coroutine which returns
    :data:`False` is substituted.

    If `no_verify` is true, none of the above regarding certificate verifiers
    matters. The internal null verifier is used, which disables certificate
    verification completely.

    .. warning::

       Disabling certificate verification makes your application vulnerable to
       trivial Man-in-the-Middle attacks. Do **not** use this outside
       controlled test environments.

       If you need to handle certificates which cannot be verified using the
       public key infrastructure, consider making use of the `pin_store`
       argument instead.

    `anonymous` may be a string or :data:`False`. If it is not :data:`False`,
    :class:`AnonymousSASLProvider` is used before password based authentication
    is attempted. In addition, it is allowed to set `password_provider` to
    :data:`None`. `anonymous` is the trace token to use, and SHOULD be the
    empty string (as specified by :xep:`175`). This requires :mod:`aiosasl` 0.3
    or newer.

    .. note::

       :data:`False` and ``""`` are treated differently for the `anonymous`
       argument, despite both being false-y values!

    .. warning::

       Take the security and privacy considerations from :rfc:`4505` (which
       specifies the ANONYMOUS SASL mechanism) and :xep:`175` (which discusses
       the ANONYMOUS SASL mechanism in the XMPP context) into account before
       using `anonymous`.

    The versaility and simplicity of use of this function make (pun intended)
    it the preferred way to construct :class:`SecurityLayer` instances.

    .. versionadded:: 0.8

       Support for SASL ANONYMOUS was added.
    """

    if isinstance(password_provider, str):
        static_password = password_provider

        @asyncio.coroutine
        def password_provider(jid, nattempt):
            if nattempt == 0:
                return static_password
            return None

    if pin_store is not None:
        if post_handshake_deferred_failure is None:
            @asyncio.coroutine
            def post_handshake_deferred_failure(verifier):
                return False

        if not isinstance(pin_store, AbstractPinStore):
            pin_data = pin_store
            if pin_type == PinType.PUBLIC_KEY:
                logger.debug("using PublicKeyPinStore")
                pin_store = PublicKeyPinStore()
            else:
                logger.debug("using CertificatePinStore")
                pin_store = CertificatePinStore()
            pin_store.import_from_json(pin_data)

        def certificate_verifier_factory():
            return PinningPKIXCertificateVerifier(
                pin_store.query,
                post_handshake_deferred_failure,
            )
    elif no_verify:
        certificate_verifier_factory = _NullVerifier
    else:
        certificate_verifier_factory = PKIXCertificateVerifier

    sasl_providers = []
    if anonymous is not False:
        if AnonymousSASLProvider is None:
            raise RuntimeError(
                "aiosasl does not support ANONYMOUS, please upgrade"
            )
        sasl_providers.append(
            AnonymousSASLProvider(anonymous)
        )

    if password_provider is not None:
        sasl_providers.append(
            PasswordSASLProvider(
                password_provider,
            ),
        )

    return SecurityLayer(
        default_ssl_context,
        certificate_verifier_factory,
        True,
        tuple(sasl_providers),
    )



########################################################################
# File name: test___init__.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import unittest

import aioxmpp.structs as structs
import aioxmpp.xso as xso


class Testtag_to_str(unittest.TestCase):
    def test_unqualified(self):
        self.assertEqual(
            "foo",
            xso.tag_to_str((None, "foo"))
        )

    def test_with_namespace(self):
        self.assertEqual(
            "{uri:bar}foo",
            xso.tag_to_str(("uri:bar", "foo"))
        )


class Testnormalize_tag(unittest.TestCase):
    def test_unqualified(self):
        self.assertEqual(
            (None, "foo"),
            xso.normalize_tag("foo")
        )

    def test_with_namespace(self):
        self.assertEqual(
            ("uri:bar", "foo"),
            xso.normalize_tag(("uri:bar", "foo"))
        )

    def test_etree_format(self):
        self.assertEqual(
            ("uri:bar", "foo"),
            xso.normalize_tag("{uri:bar}foo")
        )

    def test_validate_etree_format(self):
        with self.assertRaises(ValueError):
            xso.normalize_tag("uri:bar}foo")

    def test_validate_tuple_format(self):
        with self.assertRaises(ValueError):
            xso.normalize_tag(("foo",))
        with self.assertRaises(ValueError):
            xso.normalize_tag(("foo", "bar", "baz"))
        with self.assertRaises(ValueError):
            xso.normalize_tag(("foo", None))
        with self.assertRaises(ValueError):
            xso.normalize_tag((None, None))
        with self.assertRaises(TypeError):
            xso.normalize_tag((1, 2))

    def test_reject_incorrect_types(self):
        with self.assertRaises(TypeError):
            xso.normalize_tag(1)
        with self.assertRaises(TypeError):
            xso.normalize_tag((1, 2))


class TestAbstractTextChild(unittest.TestCase):
    def test_is_xso(self):
        self.assertTrue(issubclass(
            xso.AbstractTextChild,
            xso.XSO
        ))

    def test_has_no_tag(self):
        self.assertFalse(hasattr(xso.AbstractTextChild, "TAG"))

    def test_lang_attr(self):
        self.assertIsInstance(
            xso.AbstractTextChild.lang.xq_descriptor,
            xso.LangAttr
        )

    def test_text_attr(self):
        self.assertIsInstance(
            xso.AbstractTextChild.text.xq_descriptor,
            xso.Text
        )
        self.assertIsNone(
            xso.AbstractTextChild.text.default
        )

    def test_init_default(self):
        atc = xso.AbstractTextChild()
        self.assertIsNone(atc.lang)
        self.assertFalse(atc.text)

    def test_init_args(self):
        atc = xso.AbstractTextChild(
            "foo",
            lang=structs.LanguageTag.fromstr("de-DE"))
        self.assertEqual(atc.text, "foo")
        self.assertEqual(atc.lang, structs.LanguageTag.fromstr("de-DE"))

    def test_equality(self):
        atc1 = xso.AbstractTextChild()
        atc2 = xso.AbstractTextChild()

        self.assertTrue(atc1 == atc2)
        self.assertFalse(atc1 != atc2)

        atc1.text = "foo"

        self.assertFalse(atc1 == atc2)
        self.assertTrue(atc1 != atc2)

        atc2.text = "foo"
        atc2.lang = structs.LanguageTag.fromstr("de-DE")

        self.assertFalse(atc1 == atc2)
        self.assertTrue(atc1 != atc2)

        atc1.lang = atc2.lang

        self.assertTrue(atc1 == atc2)
        self.assertFalse(atc1 != atc2)

    def test_equality_handles_incorrect_peer_type_gracefully(self):
        atc = xso.AbstractTextChild()
        self.assertFalse(atc is None)
        self.assertFalse(atc == "foo")


class TestNO_DEFAULT(unittest.TestCase):
    def test_unequal_to_things(self):
        NO_DEFAULT = xso.NO_DEFAULT
        self.assertFalse(NO_DEFAULT is None)
        self.assertFalse(NO_DEFAULT is True)
        self.assertFalse(NO_DEFAULT is False)
        self.assertFalse(NO_DEFAULT == "")
        self.assertFalse(NO_DEFAULT == "foo")
        self.assertFalse(NO_DEFAULT == 0)
        self.assertFalse(NO_DEFAULT == 123)
        self.assertFalse(NO_DEFAULT == 0.)
        self.assertFalse(NO_DEFAULT == float("nan"))

    def test_equal_to_itself(self):
        self.assertTrue(xso.NO_DEFAULT == xso.NO_DEFAULT)
        self.assertFalse(xso.NO_DEFAULT != xso.NO_DEFAULT)

    def test___bool__raises(self):
        with self.assertRaises(TypeError):
            bool(xso.NO_DEFAULT)

    def test___index__raises(self):
        with self.assertRaises(TypeError):
            int(xso.NO_DEFAULT)

    def test___float__raises(self):
        with self.assertRaises(TypeError):
            float(xso.NO_DEFAULT)

    def test___lt__raises(self):
        with self.assertRaises(TypeError):
            xso.NO_DEFAULT < xso.NO_DEFAULT

    def test___gt__raises(self):
        with self.assertRaises(TypeError):
            xso.NO_DEFAULT > xso.NO_DEFAULT



########################################################################
# File name: upload.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import argparse
import asyncio
import configparser
import os
import pathlib
import re
import subprocess
import sys

import aiohttp

import aioxmpp
import aioxmpp.httpupload

from aioxmpp.utils import namespaces

from framework import Example, exec_example


@aiohttp.streamer
def file_sender(writer, file_, size, show_progress):
    try:
        pos = file_.tell()
    except (OSError, AttributeError):
        pos = 0

    while True:
        data = file_.read(4096)
        if not data:
            return

        pos += len(data)

        if show_progress:
            print(
                "\r{:>3.0f}%".format((pos / size) * 100),
                flush=True,
                end="",
            )

        yield from writer.write(data)


class Upload(Example):
    VALID_MIME_RE = re.compile(r"^\w+/\w+$")
    DEFAULT_MIME_TYPE = "application/octet-stream"

    def prepare_argparse(self):
        super().prepare_argparse()

        # this gives a nicer name in argparse errors
        def jid(s):
            return aioxmpp.JID.fromstr(s)

        mutex = self.argparse.add_mutually_exclusive_group()

        mutex.add_argument(
            "-s", "--service",
            default=None,
            type=jid,
            help="The HTTP Upload service to use. Omit to auto-discover."
        )

        mutex.add_argument(
            "--service-discover",
            dest="service",
            action="store_const",
            const=False,
            help="Force auto-discovery, even if a service is configured."
        )

        self.argparse.add_argument(
            "-t", "--mime-type", "--content-type",
            default=None,
            help="Content / MIME type of the file "
            "(will attempt to auto-detect if omitted)"
        )

        mutex = self.argparse.add_mutually_exclusive_group()

        mutex.add_argument(
            "--quiet",
            dest="progress",
            action="store_false",
            default=os.isatty(sys.stdout.fileno()),
            help="Do not print progress",
        )

        mutex.add_argument(
            "--progress",
            dest="progress",
            action="store_true",
            help="Print progress",
        )

        self.argparse.add_argument(
            "file",
            default=None,
            type=argparse.FileType("rb"),
            help="File to upload"
        )

    def configure(self):
        super().configure()

        self.service_addr = self.args.service
        if self.service_addr is None:
            try:
                self.service_addr = aioxmpp.JID.fromstr(
                    self.config.get("upload", "service_address")
                )
            except (configparser.NoSectionError,
                    configparser.NoOptionError):
                pass

        self.file = self.args.file
        self.file_name = pathlib.Path(self.file.name).name
        self.file_size = os.fstat(self.file.fileno()).st_size
        self.file_type = self.args.mime_type
        self.show_progress = self.args.progress

        if not self.file_type:
            try:
                self.file_type = subprocess.check_output([
                    "xdg-mime", "query", "filetype",
                    self.file.name,
                ]).decode().strip()
            except subprocess.CalledProcessError:
                self.file_type = self.DEFAULT_MIME_TYPE
                print("warning: failed to determine mime type, using {}".format(
                    self.file_type,
                ))

    @asyncio.coroutine
    def _check_for_upload_service(self, disco, jid):
        info = yield from disco.query_info(jid)
        if namespaces.xep0363_http_upload in info.features:
            return jid

    async def upload(self, url, headers):
        headers["Content-Type"] = self.file_type
        headers["Content-Length"] = str(self.file_size)
        headers["User-Agent"] = "aioxmpp/{}".format(aioxmpp.__version__)

        async with aiohttp.ClientSession() as session:
            async with session.put(
                    url,
                    data=file_sender(file_=self.file,
                                     size=self.file_size,
                                     show_progress=self.show_progress),
                    headers=headers) as response:
                if response.status not in (200, 201):
                    print(
                        "error: upload failed: {}".format(response.reason),
                        file=sys.stderr,
                    )
                    return False

        return True

    @asyncio.coroutine
    def run_simple_example(self):
        if not self.service_addr:
            disco = self.client.summon(aioxmpp.DiscoClient)
            items = yield from disco.query_items(
                self.client.local_jid.replace(localpart=None, resource=None),
                timeout=10
            )

            lookups = []

            for item in items.items:
                if item.node:
                    continue
                lookups.append(self._check_for_upload_service(disco, item.jid))

            jids = list(filter(
                None,
                (yield from asyncio.gather(*lookups))
            ))

            if not jids:
                print("error: failed to auto-discover upload service",
                      file=sys.stderr)
                return

            self.service_addr = jids[0]

        print("using {}".format(self.service_addr), file=sys.stderr)

        slot = yield from self.client.send(
            aioxmpp.IQ(
                to=self.service_addr,
                type_=aioxmpp.IQType.GET,
                payload=aioxmpp.httpupload.Request(
                    self.file_name,
                    self.file_size,
                    self.file_type,
                )
            )
        )

        if not (yield from self.upload(slot.put.url, slot.put.headers)):
            return

        print(slot.get.url)


if __name__ == "__main__":
    exec_example(Upload())



########################################################################
# File name: test_service.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import asyncio
import base64
import contextlib
import hashlib
import unittest

import aioxmpp
import aioxmpp.errors as errors
import aioxmpp.avatar.service as avatar_service
import aioxmpp.avatar.xso as avatar_xso
import aioxmpp.pubsub.xso as pubsub_xso
import aioxmpp.vcard
import aioxmpp.vcard.xso as vcard_xso

from aioxmpp.utils import namespaces
from aioxmpp.testutils import (
    make_connected_client,
    CoroutineMock,
    run_coroutine,
)

TEST_IMAGE = base64.decodebytes(b"""

""")

sha1 = hashlib.sha1()
sha1.update(TEST_IMAGE)
TEST_IMAGE_SHA1 = sha1.hexdigest()
del sha1

# jids used in the tests
TEST_FROM = aioxmpp.structs.JID.fromstr("foo@bar.example/baz")
TEST_FROM_OTHER = aioxmpp.structs.JID.fromstr("foo@bar.example/quux")
TEST_JID1 = aioxmpp.structs.JID.fromstr("bar@bar.example/baz")
TEST_JID2 = aioxmpp.structs.JID.fromstr("baz@bar.example/baz")
TEST_JID3 = aioxmpp.structs.JID.fromstr("baz@bar.example/quux")


class TestAvatarSet(unittest.TestCase):

    def test_construction(self):
        aset = avatar_service.AvatarSet()
        aset.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        aset.add_avatar_image("image/png",
                              nbytes=0,
                              id_="0000000000",
                              url="http://example.com/avatar")

        self.assertEqual(
            aset.image_bytes,
            TEST_IMAGE
        )

        self.assertEqual(
            aset.png_id,
            TEST_IMAGE_SHA1,
        )

        self.assertEqual(
            aset.metadata.info["image/png"][0].nbytes,
            len(TEST_IMAGE)
        )

        self.assertEqual(
            aset.metadata.info["image/png"][0].id_,
            TEST_IMAGE_SHA1,
        )

        self.assertEqual(
            aset.metadata.info["image/png"][1].nbytes,
            0,
        )

        self.assertEqual(
            aset.metadata.info["image/png"][1].id_,
            "0000000000",
        )

        self.assertEqual(
            aset.metadata.info["image/png"][1].url,
            "http://example.com/avatar",
        )

    def test_correct_redundant_information_is_ignored(self):
        try:
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  image_bytes=TEST_IMAGE,
                                  nbytes=len(TEST_IMAGE),
                                  id_=TEST_IMAGE_SHA1)
        except RuntimeError:
            self.fail("raises on correct redundant information")

    def test_png_id_is_normalized(self):
        for transmuted_sha1 in (TEST_IMAGE_SHA1.lower(),
                                TEST_IMAGE_SHA1.upper()):

            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  image_bytes=TEST_IMAGE,
                                  id_=transmuted_sha1)

            self.assertEqual(
                aset.metadata.info["image/png"][0].id_,
                avatar_service.normalize_id(TEST_IMAGE_SHA1),
            )

    def test_error_id_missing(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^The SHA1 of the image data is not given an not inferable "):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  nbytes=1024,
                                  url="http://example.com/avatar")

    def test_error_nbytes_missing(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^Image data length is not given an not inferable "):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  id_="00000000000000000000",
                                  url="http://example.com/avatar")

    def test_error_no_image_given(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^Either the image bytes or an url to retrieve the avatar "):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  id_="00000000000000000000",
                                  nbytes=0)

    def test_error_image_data_for_something_other_than_png(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^The image bytes can only be given for image/png data\.$"):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/gif",
                                  nbytes=1024,
                                  id_="00000000000000000000",
                                  image_bytes=TEST_IMAGE)

    def test_error_two_items_with_image_data(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^Only one avatar image may be published directly\.$"):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  image_bytes=TEST_IMAGE)
            aset.add_avatar_image("image/png",
                                  image_bytes=TEST_IMAGE)

    def test_error_redundant_sha_mismatch(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^The given id does not match the SHA1 of the image data\.$"):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  id_="00000000000000000000",
                                  image_bytes=TEST_IMAGE)

    def test_error_redundant_nbytes_mismatch(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^The given length does not match the length "
                "of the image data\.$"):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png",
                                  nbytes=0,
                                  image_bytes=TEST_IMAGE)

    def test_error_no_image(self):
        with self.assertRaisesRegex(
                RuntimeError,
                "^Either the image bytes or an url to retrieve the avatar "):
            aset = avatar_service.AvatarSet()
            aset.add_avatar_image("image/png")


class TestAvatarService(unittest.TestCase):
    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM

        self.disco_client = aioxmpp.DiscoClient(self.cc)
        self.disco = aioxmpp.DiscoServer(self.cc)
        self.presence_dispatcher = aioxmpp.dispatcher.SimplePresenceDispatcher(
            self.cc
        )
        self.presence_client = aioxmpp.PresenceClient(self.cc, dependencies={
            aioxmpp.dispatcher.SimplePresenceDispatcher:
                self.presence_dispatcher,
        })
        self.presence_server = aioxmpp.PresenceServer(self.cc)
        self.vcard = aioxmpp.vcard.VCardService(self.cc)

        self.pubsub = aioxmpp.PubSubClient(self.cc, dependencies={
            aioxmpp.DiscoClient: self.disco_client
        })

        self.pep = aioxmpp.PEPClient(self.cc, dependencies={
            aioxmpp.DiscoClient: self.disco_client,
            aioxmpp.DiscoServer: self.disco,
            aioxmpp.PubSubClient: self.pubsub,
        })

        self.s = avatar_service.AvatarService(self.cc, dependencies={
            aioxmpp.DiscoClient: self.disco_client,
            aioxmpp.DiscoServer: self.disco,
            aioxmpp.PubSubClient: self.pubsub,
            aioxmpp.PEPClient: self.pep,
            aioxmpp.PresenceClient: self.presence_client,
            aioxmpp.PresenceServer: self.presence_server,
            aioxmpp.vcard.VCardService: self.vcard
        })

        self.pep._check_for_pep = CoroutineMock()
        self.pep.available = CoroutineMock()
        self.pep.available.return_value = True

        self.cc.mock_calls.clear()

    def tearDown(self):
        del self.s
        del self.cc
        del self.disco_client
        del self.disco
        del self.pubsub

    def test_is_service(self):
        self.assertTrue(issubclass(
            avatar_service.AvatarService,
            aioxmpp.service.Service
        ))

    def test_service_order(self):
        self.assertGreater(
            avatar_service.AvatarService,
            aioxmpp.DiscoClient,
        )

        self.assertGreater(
            avatar_service.AvatarService,
            aioxmpp.DiscoServer,
        )

        self.assertGreater(
            avatar_service.AvatarService,
            aioxmpp.PubSubClient,
        )

    def test_metadata_cache_size(self):
        self.assertEqual(self.s.metadata_cache_size, 200)
        self.s.metadata_cache_size = 100
        self.assertEqual(self.s.metadata_cache_size, 100)

    def test_handle_stream_destroyed_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.stream.StanzaStream,
            "on_stream_destroyed",
            self.s.handle_stream_destroyed
        ))

    def test_handle_stream_destroyer(self):
        # for now just check the code can be run without errors,
        # it is difficult to check this properly, as we do not
        # know which caches should be wiped a priori
        self.s.handle_stream_destroyed(unittest.mock.Mock())

    def test_attach_vcard_notify_to_presence_is_depfilter(self):
        self.assertTrue(aioxmpp.service.is_depfilter_handler(
            aioxmpp.stream.StanzaStream,
            "service_outbound_presence_filter",
            self.s._attach_vcard_notify_to_presence
        ))

    def test_attach_vcard_notify_to_presence(self):
        stanza = aioxmpp.Presence()
        stanza = self.s._attach_vcard_notify_to_presence(stanza)
        self.assertIsInstance(stanza.xep0153_x,
                              avatar_xso.VCardTempUpdate)
        self.assertIsNone(stanza.xep0153_x.photo)

        vcard_id = '1234'
        self.s._vcard_id = vcard_id
        stanza = aioxmpp.Presence()
        stanza = self.s._attach_vcard_notify_to_presence(stanza)
        self.assertIsInstance(stanza.xep0153_x,
                              avatar_xso.VCardTempUpdate)
        self.assertEqual(stanza.xep0153_x.photo, vcard_id)

        self.s._vcard_resource_interference.add(TEST_JID1)
        stanza = aioxmpp.Presence()
        stanza = self.s._attach_vcard_notify_to_presence(stanza)
        self.assertIsNotNone(stanza.xep0153_x)
        self.assertIsNone(stanza.xep0153_x.photo)

    def test_handle_on_available_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.PresenceClient,
            "on_available",
            self.s._handle_on_available
        ))

    def test_resource_interference(self):
        stanza = aioxmpp.Presence()

        self.s._handle_on_available(TEST_FROM_OTHER, stanza)
        self.assertCountEqual(
            self.s._vcard_resource_interference,
            [TEST_FROM_OTHER]
        )

        stanza = aioxmpp.Presence()
        self.s._handle_on_unavailable(TEST_FROM_OTHER, stanza)
        self.assertCountEqual(
            self.s._vcard_resource_interference,
            []
        )

    def test_trigger_rehash(self):
        mock_handler = unittest.mock.Mock()
        self.s.on_metadata_changed.connect(mock_handler)

        stanza = aioxmpp.Presence()
        stanza.xep0153_x = avatar_xso.VCardTempUpdate("1234")
        self.s._handle_on_available(TEST_FROM_OTHER, stanza)

        first_rehash_task = self.s._vcard_rehash_task
        self.assertIsNot(first_rehash_task, None)

        # presence with the same hash does not affect the rehash task
        self.s._vcard_id = "1234"
        stanza = aioxmpp.Presence()
        stanza.xep0153_x = avatar_xso.VCardTempUpdate("1234")
        self.s._handle_on_available(TEST_FROM_OTHER, stanza)
        self.assertIs(self.s._vcard_rehash_task, first_rehash_task)

        # presence with another hash cancels the task
        stanza = aioxmpp.Presence()
        stanza.xep0153_x = avatar_xso.VCardTempUpdate("4321")
        self.s._handle_on_available(TEST_FROM_OTHER, stanza)

        with unittest.mock.patch.object(self.vcard, "get_vcard",
                                        new=CoroutineMock()):
            vcard = vcard_xso.VCard()
            self.vcard.get_vcard.return_value = vcard
            vcard.set_photo_data("image/png", TEST_IMAGE)

            loop = asyncio.get_event_loop()
            with self.assertRaises(asyncio.CancelledError):
                loop.run_until_complete(first_rehash_task)
            self.assertTrue(first_rehash_task.cancelled())

            loop.run_until_complete(
                self.s._vcard_rehash_task
            )

        self.assertEqual(self.s._vcard_id, TEST_IMAGE_SHA1)

        with contextlib.ExitStack() as stack:
            stack.enter_context(
                unittest.mock.patch.object(self.vcard, "get_vcard",
                                           new=CoroutineMock())
            )

            resend_presence = stack.enter_context(
                unittest.mock.patch.object(
                    self.presence_server,
                    "resend_presence",
                )
            )

            vcard = vcard_xso.VCard()
            self.vcard.get_vcard.return_value = vcard

            stanza = aioxmpp.Presence()
            stanza.xep0153_x = avatar_xso.VCardTempUpdate("9132752")
            self.s._handle_on_available(TEST_FROM_OTHER, stanza)

            loop.run_until_complete(
                self.s._vcard_rehash_task
            )

        resend_presence.assert_called_once_with()

        self.assertEqual(self.s._vcard_id, "")

        # XXX: should we test minutely that we get the right metadata
        mock_handler.assert_called_with(TEST_FROM_OTHER.bare(),
                                        unittest.mock.ANY)

    def test_handle_on_changed_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.PresenceClient,
            "on_changed",
            self.s._handle_on_changed
        ))

    def test_handle_notify_without_photo_is_noop(self):
        mock_handler = unittest.mock.Mock()
        self.s.on_metadata_changed.connect(mock_handler)

        stanza = aioxmpp.Presence()
        self.s._handle_notify(TEST_JID1, stanza)
        stanza.xep0153_x = avatar_xso.VCardTempUpdate()
        self.s._handle_notify(TEST_JID1, stanza)

        self.assertEqual(len(mock_handler.mock_calls), 0)

    def test_handle_on_changed(self):
        with unittest.mock.patch.object(self.s, "_handle_notify"):
            self.s._handle_on_changed(unittest.mock.sentinel.jid,
                                      unittest.mock.sentinel.staza)
            self.assertSequenceEqual(
                self.s._handle_notify.mock_calls,
                [
                    unittest.mock.call(unittest.mock.sentinel.jid,
                                       unittest.mock.sentinel.staza)
                ]
            )

    def test_handle_on_unavailable_is_depsignal_handler(self):
        self.assertTrue(aioxmpp.service.is_depsignal_handler(
            aioxmpp.PresenceClient,
            "on_unavailable",
            self.s._handle_on_unavailable
        ))

    def test_publish_avatar_set(self):
        # set the cache to indicate the server has PEP

        avatar_set = avatar_service.AvatarSet()
        avatar_set.add_avatar_image("image/png", image_bytes=TEST_IMAGE)

        with unittest.mock.patch.object(self.pep, "publish",
                                        new=CoroutineMock()):
            run_coroutine(self.s.publish_avatar_set(avatar_set))

            self.assertSequenceEqual(
                self.pep.publish.mock_calls,
                [
                    unittest.mock.call(
                        namespaces.xep0084_data,
                        unittest.mock.ANY,
                        id_=avatar_set.png_id
                    ),
                    unittest.mock.call(
                        namespaces.xep0084_metadata,
                        avatar_set.metadata,
                        id_=avatar_set.png_id
                    )
                ],
            )

            _, args, _ = self.pep.publish.mock_calls[0]
            data = args[1]
            self.assertTrue(isinstance(data, avatar_xso.Data))
            self.assertEqual(data.data, avatar_set.image_bytes)

    def test_publish_avatar_no_protocol_raises(self):
        self.pep.available.return_value = False
        self.s.synchronize_vcard = False

        with self.assertRaises(RuntimeError):
            run_coroutine(self.s.publish_avatar_set(unittest.mock.Mock()))

    def test_publish_avatar_set_synchronize_vcard(self):

        avatar_set = avatar_service.AvatarSet()
        avatar_set.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        self.assertFalse(self.s.synchronize_vcard)
        self.s.synchronize_vcard = True
        self.assertTrue(self.s.synchronize_vcard)

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(self.pep, "publish",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.presence_server,
                                                       "resend_presence"))
            e.enter_context(unittest.mock.patch.object(self.vcard, "get_vcard",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.vcard, "set_vcard",
                                                       new=CoroutineMock()))
            self.vcard.get_vcard.return_value = unittest.mock.Mock()
            run_coroutine(self.s.publish_avatar_set(avatar_set))

            self.assertSequenceEqual(
                self.presence_server.resend_presence.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(
                self.vcard.get_vcard.mock_calls,
                [
                    unittest.mock.call(),
                    unittest.mock.call().set_photo_data("image/png",
                                                        TEST_IMAGE),
                ]
            )

            self.assertSequenceEqual(
                self.vcard.set_vcard.mock_calls,
                [
                    unittest.mock.call(self.vcard.get_vcard.return_value),
                ]
            )

            self.assertSequenceEqual(
                self.pep.publish.mock_calls,
                [
                    unittest.mock.call(
                        namespaces.xep0084_data,
                        unittest.mock.ANY,
                        id_=avatar_set.png_id
                    ),
                    unittest.mock.call(
                        namespaces.xep0084_metadata,
                        avatar_set.metadata,
                        id_=avatar_set.png_id
                    )
                ],
            )

            _, args, _ = self.pep.publish.mock_calls[0]
            data = args[1]
            self.assertTrue(isinstance(data, avatar_xso.Data))
            self.assertEqual(data.data, avatar_set.image_bytes)

    def test_publish_avatar_set_synchronize_vcard_pep_raises(self):
        avatar_set = avatar_service.AvatarSet()
        avatar_set.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        self.s.synchronize_vcard = True

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(self.pep, "publish",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.presence_server,
                                                       "resend_presence"))
            e.enter_context(unittest.mock.patch.object(self.vcard, "get_vcard",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.vcard, "set_vcard",
                                                       new=CoroutineMock()))

            # do not do the vcard operations of pep is available but
            # fails
            self.pep.publish.side_effect = RuntimeError

            self.vcard.get_vcard.return_value = unittest.mock.Mock()

            with self.assertRaises(RuntimeError):
                run_coroutine(self.s.publish_avatar_set(avatar_set))

            self.assertSequenceEqual(
                self.presence_server.resend_presence.mock_calls,
                []
            )

            self.assertSequenceEqual(
                self.vcard.get_vcard.mock_calls,
                []
            )

            self.assertSequenceEqual(
                self.vcard.set_vcard.mock_calls,
                []
            )

    def test_disable_avatar(self):
        with unittest.mock.patch.object(self.pep, "publish",
                                        new=CoroutineMock()):
            run_coroutine(self.s.disable_avatar())

            self.assertSequenceEqual(
                self.pep.publish.mock_calls,
                [
                    unittest.mock.call(
                        namespaces.xep0084_metadata,
                        unittest.mock.ANY,
                    ),
                ]
            )

            _, args, _ = self.pep.publish.mock_calls[0]
            metadata = args[1]

            self.assertTrue(isinstance(metadata, avatar_xso.Metadata))
            self.assertEqual(0, len(metadata.info))
            self.assertEqual(0, len(metadata.pointer))

    def test_wipe_avatar(self):
        with unittest.mock.patch.object(self.pep, "publish",
                                        new=CoroutineMock()):
            run_coroutine(self.s.wipe_avatar())

            self.assertSequenceEqual(
                self.pep.publish.mock_calls,
                [
                    unittest.mock.call(
                        namespaces.xep0084_metadata,
                        unittest.mock.ANY,
                    ),
                    unittest.mock.call(
                        namespaces.xep0084_data,
                        unittest.mock.ANY,
                    ),
                ]
            )

            _, args, _ = self.pep.publish.mock_calls[0]
            metadata = args[1]

            self.assertTrue(isinstance(metadata, avatar_xso.Metadata))
            self.assertEqual(0, len(metadata.info))
            self.assertEqual(0, len(metadata.pointer))

            _, args, _ = self.pep.publish.mock_calls[1]
            data = args[1]

            self.assertTrue(isinstance(data, avatar_xso.Data))
            self.assertEqual(0, len(data.data))

    def test_wipe_avatar_with_vcard(self):
        self.s.synchronize_vcard = True
        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(self.pep, "publish",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.presence_server,
                                                       "resend_presence"))
            e.enter_context(unittest.mock.patch.object(self.vcard, "get_vcard",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.vcard, "set_vcard",
                                                       new=CoroutineMock()))
            self.vcard.get_vcard.return_value = unittest.mock.Mock()
            run_coroutine(self.s.wipe_avatar())

            self.assertSequenceEqual(
                self.presence_server.resend_presence.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(
                self.vcard.get_vcard.mock_calls,
                [unittest.mock.call(),
                 unittest.mock.call().clear_photo_data()]
            )

            self.assertSequenceEqual(
                self.vcard.set_vcard.mock_calls,
                [unittest.mock.call(unittest.mock.ANY)]
            )

            self.assertSequenceEqual(
                self.pep.publish.mock_calls,
                [
                    unittest.mock.call(
                        namespaces.xep0084_metadata,
                        unittest.mock.ANY,
                    ),
                    unittest.mock.call(
                        namespaces.xep0084_data,
                        unittest.mock.ANY,
                    ),
                ]
            )

            _, args, _ = self.pep.publish.mock_calls[0]
            metadata = args[1]

            self.assertTrue(isinstance(metadata, avatar_xso.Metadata))
            self.assertEqual(0, len(metadata.info))
            self.assertEqual(0, len(metadata.pointer))

            _, args, _ = self.pep.publish.mock_calls[1]
            data = args[1]

            self.assertTrue(isinstance(data, avatar_xso.Data))
            self.assertEqual(0, len(data.data))

    def test_disable_avatar_synchronize_vcard_pep_raises(self):
        self.s.synchronize_vcard = True

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(self.pep, "publish",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.presence_server,
                                                       "resend_presence"))
            e.enter_context(unittest.mock.patch.object(self.vcard, "get_vcard",
                                                       new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(self.vcard, "set_vcard",
                                                       new=CoroutineMock()))

            # do not do the vcard operations of pep is available but
            # fails
            self.pep.publish.side_effect = RuntimeError

            self.vcard.get_vcard.return_value = unittest.mock.Mock()

            with self.assertRaises(RuntimeError):
                run_coroutine(self.s.disable_avatar())

            self.assertSequenceEqual(
                self.presence_server.resend_presence.mock_calls,
                [unittest.mock.call()]
            )

            self.assertSequenceEqual(
                self.vcard.get_vcard.mock_calls,
                [unittest.mock.call(),
                 unittest.mock.call().clear_photo_data()]
            )

            self.assertSequenceEqual(
                self.vcard.set_vcard.mock_calls,
                [unittest.mock.call(unittest.mock.ANY)]
            )

    def test_handle_pubsub_publish(self):
        self.assertTrue(aioxmpp.service.is_attrsignal_handler(
            avatar_service.AvatarService.avatar_pep,
            "on_item_publish",
            self.s._handle_pubsub_publish
        ))

        aset = avatar_service.AvatarSet()
        aset.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        aset.add_avatar_image("image/png",
                              nbytes=0,
                              id_="00000000000000000000",
                              url="http://example.com/avatar")

        # construct the proper pubsub response
        item = pubsub_xso.EventItem(aset.metadata, id_=aset.png_id)

        mock_handler = unittest.mock.Mock()
        self.s.on_metadata_changed.connect(mock_handler)

        self.s._handle_pubsub_publish(
            TEST_JID1,
            namespaces.xep0084_metadata,
            item)

        descriptors = self.s._metadata_cache[TEST_JID1]
        self.assertEqual(len(descriptors), 2)

        png_descr = descriptors

        self.assertTrue(isinstance(png_descr[0],
                                   avatar_service.PubsubAvatarDescriptor))
        self.assertEqual(png_descr[0].mime_type, "image/png")
        self.assertEqual(png_descr[0].id_, TEST_IMAGE_SHA1)
        self.assertEqual(png_descr[0].nbytes, len(TEST_IMAGE))
        self.assertEqual(png_descr[0].url, None)

        self.assertTrue(isinstance(png_descr[0],
                                   avatar_service.PubsubAvatarDescriptor))

        self.assertEqual(png_descr[1].mime_type, "image/png")
        self.assertEqual(png_descr[1].id_, "00000000000000000000")
        self.assertEqual(png_descr[1].nbytes, 0)
        self.assertEqual(png_descr[1].url, "http://example.com/avatar")

        mock_handler.assert_called_with(TEST_JID1, descriptors)

    def test_get_avatar_metadata(self):
        aset = avatar_service.AvatarSet()
        aset.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        aset.add_avatar_image("image/png",
                              nbytes=0,
                              id_="00000000000000000000",
                              url="http://example.com/avatar")

        # construct the proper pubsub response
        items = pubsub_xso.Items(
            namespaces.xep0084_metadata,
        )
        item = pubsub_xso.Item(id_=aset.png_id)
        item.registered_payload = aset.metadata
        items.items.append(item)

        pubsub_result = pubsub_xso.Request(items)

        with unittest.mock.patch.object(self.pubsub, "get_items",
                                        new=CoroutineMock()):
            self.pubsub.get_items.return_value = pubsub_result

            descriptors = run_coroutine(self.s.get_avatar_metadata(TEST_JID1))

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_metadata,
                        max_items=1
                    )
                ]
            )

            self.assertEqual(len(descriptors), 2)

            png_descr = descriptors

            self.assertTrue(isinstance(png_descr[0],
                                       avatar_service.PubsubAvatarDescriptor))
            self.assertEqual(png_descr[0].mime_type, "image/png")
            self.assertEqual(png_descr[0].id_, TEST_IMAGE_SHA1)
            self.assertEqual(png_descr[0].nbytes, len(TEST_IMAGE))
            self.assertEqual(png_descr[0].url, None)

            self.assertTrue(isinstance(png_descr[0],
                                       avatar_service.PubsubAvatarDescriptor))

            self.assertEqual(png_descr[1].mime_type, "image/png")
            self.assertEqual(png_descr[1].id_, "00000000000000000000")
            self.assertEqual(png_descr[1].nbytes, 0)
            self.assertEqual(png_descr[1].url, "http://example.com/avatar")

        with unittest.mock.patch.object(self.pubsub, "get_items",
                                        new=CoroutineMock()):

            cached_descriptors = run_coroutine(
                self.s.get_avatar_metadata(TEST_JID1)
            )

            self.assertEqual(descriptors, cached_descriptors)

            # we must get the descriptors from the cache
            self.pubsub.get_items.assert_not_called()

        with unittest.mock.patch.object(self.pubsub, "get_items",
                                        new=CoroutineMock()):

            self.pubsub.get_items.return_value = pubsub_result
            descriptors = run_coroutine(
                self.s.get_avatar_metadata(TEST_JID1, require_fresh=True)
            )

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_metadata,
                        max_items=1
                    )
                ]
            )

    def test_get_avatar_metadata_with_require_fresh_does_not_crash(self):
        aset = avatar_service.AvatarSet()
        aset.add_avatar_image("image/png", image_bytes=TEST_IMAGE)
        aset.add_avatar_image("image/png",
                              nbytes=0,
                              id_="00000000000000000000",
                              url="http://example.com/avatar")

        # construct the proper pubsub response
        items = pubsub_xso.Items(
            namespaces.xep0084_metadata,
        )
        item = pubsub_xso.Item(id_=aset.png_id)
        item.registered_payload = aset.metadata
        items.items.append(item)

        pubsub_result = pubsub_xso.Request(items)

        with unittest.mock.patch.object(self.pubsub, "get_items",
                                        new=CoroutineMock()):

            self.pubsub.get_items.return_value = pubsub_result
            run_coroutine(
                self.s.get_avatar_metadata(TEST_JID1, require_fresh=True)
            )

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_metadata,
                        max_items=1
                    )
                ]
            )

    def test_get_avatar_metadata_vcard_fallback(self):
        sha1 = hashlib.sha1(b'')
        empty_sha1 = sha1.hexdigest()

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(
                self.pubsub, "get_items",
                new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(
                self.vcard, "get_vcard",
                new=CoroutineMock()))
            vcard_mock = unittest.mock.Mock()
            vcard_mock.get_photo_data.return_value = b''
            self.vcard.get_vcard.return_value = vcard_mock

            self.pubsub.get_items.side_effect = errors.XMPPCancelError(
                errors.ErrorCondition.FEATURE_NOT_IMPLEMENTED
            )

            res = run_coroutine(self.s.get_avatar_metadata(TEST_JID1))

            self.assertEqual(len(res), 1)
            self.assertIsInstance(res[0],
                                  avatar_service.VCardAvatarDescriptor)
            self.assertEqual(res[0]._image_bytes,
                             b'')
            self.assertEqual(res[0].id_,
                             empty_sha1)
            self.assertEqual(res[0].nbytes,
                             0)

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_metadata,
                        max_items=1
                    ),
                ]
            )

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(
                self.pubsub, "get_items",
                new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(
                self.vcard, "get_vcard",
                new=CoroutineMock()))
            vcard_mock = unittest.mock.Mock()
            vcard_mock.get_photo_data.return_value = b''
            self.vcard.get_vcard.return_value = vcard_mock

            self.pubsub.get_items.side_effect = errors.XMPPCancelError(
                errors.ErrorCondition.ITEM_NOT_FOUND
            )

            res = run_coroutine(self.s.get_avatar_metadata(TEST_JID2))

            self.assertEqual(len(res), 1)
            self.assertIsInstance(res[0],
                                  avatar_service.VCardAvatarDescriptor)
            self.assertEqual(res[0]._image_bytes,
                             b'')
            self.assertEqual(res[0].id_,
                             empty_sha1)
            self.assertEqual(res[0].nbytes,
                             0)

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID2,
                        namespaces.xep0084_metadata,
                        max_items=1
                    ),
                ]
            )

        with contextlib.ExitStack() as e:
            e.enter_context(unittest.mock.patch.object(
                self.pubsub, "get_items",
                new=CoroutineMock()))
            e.enter_context(unittest.mock.patch.object(
                self.vcard, "get_vcard",
                new=CoroutineMock()))
            vcard_mock = unittest.mock.Mock()
            vcard_mock.get_photo_data.return_value = None
            self.vcard.get_vcard.return_value = vcard_mock

            self.pubsub.get_items.side_effect = errors.XMPPCancelError(
                errors.ErrorCondition.ITEM_NOT_FOUND
            )

            res = run_coroutine(self.s.get_avatar_metadata(TEST_JID3))
            self.assertEqual(len(res), 0)

            self.assertSequenceEqual(
                self.pubsub.get_items.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID3,
                        namespaces.xep0084_metadata,
                        max_items=1
                    ),
                ]
            )

    def test_subscribe(self):
        with unittest.mock.patch.object(self.pubsub, "subscribe",
                                        new=CoroutineMock()):
            run_coroutine(self.s.subscribe(TEST_JID1))

            self.assertSequenceEqual(
                self.pubsub.subscribe.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_metadata
                    ),
                ]
            )


class TestAvatarDescriptors(unittest.TestCase):

    def setUp(self):
        self.cc = make_connected_client()
        self.cc.local_jid = TEST_FROM
        self.vcard = aioxmpp.vcard.VCardService(self.cc)
        self.disco = aioxmpp.DiscoClient(self.cc)
        self.pubsub = aioxmpp.PubSubClient(self.cc, dependencies={
            aioxmpp.DiscoClient: self.disco
        })

        self.cc.mock_calls.clear()

    def test_attributes_defined_by_AbstractAvatarDescriptor(self):
        a = avatar_service.AbstractAvatarDescriptor(
            TEST_JID1, TEST_IMAGE_SHA1, mime_type="image/png",
            nbytes=len(TEST_IMAGE)
        )
        with self.assertRaises(NotImplementedError):
            run_coroutine(a.get_image_bytes())

        self.assertFalse(a.has_image_data_in_pubsub)
        self.assertFalse(a.can_get_image_bytes_via_xmpp)

        self.assertEqual(a.remote_jid, TEST_JID1)
        self.assertEqual(a.mime_type, "image/png")
        self.assertEqual(a.id_, TEST_IMAGE_SHA1)
        self.assertEqual(a.nbytes, len(TEST_IMAGE))
        self.assertEqual(a.normalized_id,
                         avatar_service.normalize_id(TEST_IMAGE_SHA1))
        self.assertEqual(a.url, None)
        self.assertEqual(a.width, None)
        self.assertEqual(a.height, None)

    def test_avatar_descriptor_equality(self):
        vcard_descriptor = avatar_service.VCardAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            nbytes=len(TEST_IMAGE),
            vcard=self.vcard,
            image_bytes=TEST_IMAGE,
        )

        pep_descriptor = avatar_service.PubsubAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            mime_type="image/png",
            nbytes=len(TEST_IMAGE),
            pubsub=self.pubsub
        )

        pep2_descriptor = avatar_service.PubsubAvatarDescriptor(
            TEST_JID2,
            TEST_IMAGE_SHA1.upper(),
            mime_type="image/png",
            nbytes=len(TEST_IMAGE),
            pubsub=self.pubsub
        )

        url_descriptor = avatar_service.HttpAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            mime_type="image/png",
            nbytes=len(TEST_IMAGE),
            url="http://example.com/avatar"
        )

        self.assertEqual(pep_descriptor, pep_descriptor)
        self.assertEqual(pep2_descriptor, pep2_descriptor)
        self.assertEqual(url_descriptor, url_descriptor)
        self.assertEqual(vcard_descriptor, vcard_descriptor)
        self.assertNotEqual(pep_descriptor, pep2_descriptor)
        self.assertNotEqual(pep_descriptor, url_descriptor)
        self.assertNotEqual(pep_descriptor, vcard_descriptor)
        self.assertNotEqual(url_descriptor, vcard_descriptor)


    def test_vcard_get_image_bytes(self):
        descriptor = avatar_service.VCardAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            nbytes=len(TEST_IMAGE),
            vcard=self.vcard,
            image_bytes=TEST_IMAGE,
        )

        self.assertTrue(descriptor.has_image_data_in_pubsub)
        self.assertTrue(descriptor.can_get_image_bytes_via_xmpp)

        with unittest.mock.patch.object(self.vcard, "get_vcard",
                                        new=CoroutineMock()):
            self.assertEqual(
                run_coroutine(descriptor.get_image_bytes()),
                TEST_IMAGE
            )
            self.assertSequenceEqual(self.vcard.get_vcard.mock_calls, [])

        descriptor = avatar_service.VCardAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            nbytes=len(TEST_IMAGE),
            vcard=self.vcard,
        )

        with unittest.mock.patch.object(self.vcard, "get_vcard",
                                        new=CoroutineMock()):
            vcard_mock = unittest.mock.Mock()
            vcard_mock.get_photo_data.return_value = TEST_IMAGE
            self.vcard.get_vcard.return_value = vcard_mock
            self.assertEqual(
                run_coroutine(descriptor.get_image_bytes()),
                TEST_IMAGE
            )

            vcard_mock.get_photo_data.return_value = None
            with self.assertRaises(RuntimeError):
                run_coroutine(descriptor.get_image_bytes())

            self.assertSequenceEqual(
                self.vcard.get_vcard.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1
                    ),
                    unittest.mock.call().get_photo_data(),
                    unittest.mock.call(
                        TEST_JID1
                    ),
                    unittest.mock.call().get_photo_data(),
                ]
            )

    def test_pep_get_image_bytes(self):
        descriptor = avatar_service.PubsubAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            mime_type="image/png",
            nbytes=len(TEST_IMAGE),
            pubsub=self.pubsub
        )

        self.assertTrue(descriptor.has_image_data_in_pubsub)
        self.assertEqual(TEST_IMAGE_SHA1, descriptor.normalized_id)

        items = pubsub_xso.Items(
            namespaces.xep0084_data,
        )
        pubsub_result = pubsub_xso.Request(items)

        with unittest.mock.patch.object(self.pubsub, "get_items_by_id",
                                        new=CoroutineMock()):
            self.pubsub.get_items_by_id.return_value = pubsub_result
            with self.assertRaises(RuntimeError):
                res = run_coroutine(descriptor.get_image_bytes())

            item = pubsub_xso.Item(id_=TEST_IMAGE_SHA1)
            item.registered_payload = avatar_xso.Data(TEST_IMAGE)
            items.items.append(item)

            res = run_coroutine(descriptor.get_image_bytes())

            self.assertSequenceEqual(
                self.pubsub.get_items_by_id.mock_calls,
                [
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_data,
                        [TEST_IMAGE_SHA1.upper()],
                    ),
                    unittest.mock.call(
                        TEST_JID1,
                        namespaces.xep0084_data,
                        [TEST_IMAGE_SHA1.upper()],
                    )
                ]
            )
            self.assertEqual(res, TEST_IMAGE)

    def test_HttpAvatarDescriptor(self):
        descriptor = avatar_service.HttpAvatarDescriptor(
            TEST_JID1,
            TEST_IMAGE_SHA1.upper(),
            mime_type="image/png",
            nbytes=len(TEST_IMAGE),
            url="http://example.com/avatar"
        )

        self.assertFalse(descriptor.has_image_data_in_pubsub)
        with self.assertRaises(NotImplementedError):
            run_coroutine(descriptor.get_image_bytes())



#!/usr/bin/env python3
########################################################################
# File name: log-to-scs.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
########################################################################
import ast
import enum
import re

import lxml.etree as etree

import aioxmpp


_client_prefix = r"^aioxmpp\.e2etest\.provision\.client(?P<client_id>[0-9]+)"


SENT_LINE = re.compile(
    _client_prefix + r"\.XMLStream:\ DEBUG:\ SENT\ (?P<bytes>b.+)$",
    re.VERBOSE
)


RECV_LINE = re.compile(
    _client_prefix + r"\.XMLStream:\ DEBUG:\ RECV\ (?P<bytes>b.+)$",
    re.VERBOSE
)

JID_LINE = re.compile(
    _client_prefix + r":\ INFO:\ bound\ to\ jid:\ (?P<jid>.+)",
    re.VERBOSE,
)


def _parse_stream_bytes(b):
    FOOTER = b"</stream:stream>"
    b = ast.literal_eval(b)
    # stream header, ignore
    if b.startswith(b"<?xml"):
        return
    if b.startswith(FOOTER):
        return
    if b.endswith(FOOTER):
        b = b[:-len(FOOTER)]
    if not b.strip():
        return
    try:
        tree = etree.fromstring(b"<root xmlns:stream='http://etherx.jabber.org/streams'>"+b+b"</root>")
    except:
        print(b)
        raise
    yield from tree


def _wrap_bytes(match, wrapper_name):
    client_id = int(match["client_id"])
    for piece in _parse_stream_bytes(match["bytes"]):
        yield {
            "client_id": client_id,
            wrapper_name: {
                "xml": piece
            }
        }


def parse_sent_line(match):
    yield from _wrap_bytes(match, "sent")


def parse_recv_line(match):
    yield from _wrap_bytes(match, "recv")


def parse_jid_line(match):
    yield {
        "client_id": int(match["client_id"]),
        "bound": {
            "jid": aioxmpp.JID.fromstr(match["jid"]),
        }
    }


line_parsers = [
    (SENT_LINE, parse_sent_line),
    (RECV_LINE, parse_recv_line),
    (JID_LINE, parse_jid_line),
]


def parse_line(l):
    for rx, parser in line_parsers:
        match = rx.match(l)
        if match is None:
            continue
        yield from parser(match.groupdict())


def parse_lines(ls):
    for line in ls:
        parsed = parse_line(line)
        if parsed is not None:
            yield from parsed


def xmllines(tree):
    serialised = etree.tostring(tree, encoding="utf-8", pretty_print=True)
    lines = serialised.decode("utf-8").split("\n")
    return [line for line in lines if line.strip()]


def filter_sessions(actions):
    ids_to_drop = set()
    for action in actions:
        xml = action.get("sent", action.get("recv", {})).get("xml")
        if xml is None:
            yield action
            continue

        id_ = xml.get("id")
        key = (action["client_id"], id_)
        if key in ids_to_drop:
            ids_to_drop.discard(key)
            continue

        if xml.tag == "iq" and len(xml) > 0:
            # IQ with payload
            if xml[0].tag == "{urn:ietf:params:xml:ns:xmpp-session}session":
                # drop!
                ids_to_drop.add(key)
                continue

        yield action


def filter_serverdisco(actions):
    ids_to_drop = set()
    client_jids = {}
    for action in actions:
        try:
            jid = action["bound"]["jid"]
        except KeyError:
            pass
        else:
            client_jids[action["client_id"]] = jid

        xml = action.get("sent", action.get("recv", {})).get("xml")
        if xml is None:
            yield action
            continue

        id_ = xml.get("id")
        key = (action["client_id"], id_)
        if key in ids_to_drop:
            ids_to_drop.discard(key)
            continue

        try:
            client_jid = client_jids[action["client_id"]]
        except KeyError:
            yield action
            continue

        if (xml.tag == "iq" and len(xml) > 0 and
                xml.get("to") == client_jid.domain and
                xml[0].tag == "{http://jabber.org/protocol/disco#info}query"):
            # drop!
            ids_to_drop.add(key)
            continue

        yield action


class FinalAction(enum.Enum):
    CONNECT = "connects"
    SEND = "sends"
    RECEIVE = "receives"


def generate(actions, out, characters=[], remove_clients=[]):
    client_names = characters or [
        "Juliet",
        "Romeo",
    ]

    clients = {}
    result_actions = []

    def bind_client(client_id, jid):
        clients[client_id] = {
            "id": client_id,
            "name": client_names.pop(0),
            "jid": jid,
        }
        result_actions.append(
            (client_id, (FinalAction.CONNECT, None))
        )

    for action in actions:
        try:
            client = clients[action["client_id"]]
        except KeyError:
            if "bound" in action and action["client_id"] not in remove_clients:
                bind_client(action["client_id"], action["bound"]["jid"])
            continue

        if "sent" in action:
            result_actions.append(
                (client["id"], (FinalAction.SEND, action["sent"]["xml"]))
            )
        elif "recv" in action:
            result_actions.append(
                (client["id"], (FinalAction.RECEIVE, action["recv"]["xml"]))
            )
        else:
            raise RuntimeError("unknown action: {}".format(action))

    for client in sorted(clients.values(), key=lambda x: x["id"]):
        print(
            "[Client] {name}\n\tjid: {jid}\n\tpassword: password\n".format(
                **client,
            ),
            file=out,
        )

    print("---------\n")

    for client_id, (action, xml) in result_actions:
        client = clients[client_id]
        print(
            "{client[name]} {action.value}".format(
                client=client, action=action
            ),
            end="",
            file=out
        )
        if xml is not None:
            print(":\n\t{}".format("\n\t".join(xmllines(xml))), file=out)
        else:
            print(file=out)
        print(file=out)


if __name__ == "__main__":
    import argparse
    import sys

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--characters",
        nargs="+",
        dest="characters",
        default=[],
    )
    parser.add_argument(
        "--strip-serverdisco",
        action="append_const",
        dest="add_filters",
        default=[],
        const=filter_serverdisco,
    )
    parser.add_argument(
        "--no-strip-sessions",
        action="append_const",
        dest="remove_filters",
        default=[],
        const=filter_sessions,
    )
    parser.add_argument(
        "--remove-client",
        action="append",
        dest="remove_clients",
        default=[],
        type=int,
    )

    parser.add_argument(
        "-o", "--output",
        default=sys.stdout,
        type=argparse.FileType("w"),
    )

    parser.add_argument(
        "input",
        nargs="?",
        default=sys.stdin,
        type=argparse.FileType("r"),
    )

    args = parser.parse_args()

    filters = [filter_sessions]
    for to_add in args.add_filters:
        filters.append(to_add)
    for to_remove in args.remove_filters:
        try:
            filters.remove(to_remove)
        except ValueError:
            pass

    with args.input as f:
        actions = list(parse_lines(f))

    for filter_func in filters:
        actions = filter_func(actions)

    with args.output as f:
        generate(actions, f, characters=list(args.characters),
                 remove_clients=args.remove_clients)

